#
#  peppercornenumerator/reactions.py
#  EnumeratorProject
#
from __future__ import absolute_import, print_function, division
from builtins import map

import logging
log = logging.getLogger(__name__)

import math

from peppercornenumerator.utils import Loop, wrap
from peppercornenumerator.objects import DSDDuplicationError
from peppercornenumerator.objects import PepperComplex, PepperReaction 
from peppercornenumerator.objects import make_pair_table, pair_table_to_dot_bracket

# Rate constant formulas
# ----------------------------------------------------------------------------

class PeppercornRateModelError(Exception):
    pass

def opening_rate(length, dG_bp = -1.7, dG_assoc = 1.9, kelvin = 298.15, dissoc = True):
    """ Rate constant for opening a duplex of a given `length`.

    k_open = k_bind * exp( (length * dG_bp + dG_assoc) / RT )

    Args: 
        dG_bp (flt): Energy contribution of a single WC base pair. 
            Defaults to -1.7 kcal/mol.
        dG_assoc (flt): Association energy penalty.
            Defaults to 1.9 kcal/mol.
        kelvin (flt): Temperature in K. Defaults to 298.15 (25 C).

    Returns:
        [float]: The rate constant.
    """
    # R = 0.001987 kcal/mol/K
    # T = (273.15 + 25) K
    RT = 0.001987 * kelvin

    if True or dissoc: # stay with simple model!
        k_bind21 = bimolecular_binding_rate(length)
        dG = (length * dG_bp) + dG_assoc
        rate = k_bind21 * math.exp(dG/RT)
    else:
        k_bind11 = 1e6 * length # using zipping rate, not bubble-closing
        dG = length * dG_bp
        rate = k_bind11 * math.exp(dG/RT)
    return rate

def polymer_link_length(before, after):
    """ Length estimate [nt] for linkers loops connecting two domains.

    Args:
        before (Loop): A loop object for which we measure the length.
        after (Loop): A loop object for which we measure the length.

    Returns:
        [float]: The length of the shorter loop.
    """
    L_stem = 2.0 / 0.43  # rough equivalent number of single-stranded nucleotides to span a stem

    L_before = float('inf') if before.is_open else \
            1 + before.bases + before.stems + L_stem * before.stems 

    L_after = float('inf') if after.is_open else \
            1 + after.bases + after.stems + L_stem * after.stems

    if after.is_open and before.is_open:
        raise PeppercornRateModelError("Error: computing polymer lengths in disconnected complex!")

    return min(L_before, L_after)

def polymer_link_rate(hllen, ha = 1e6, hb = -2.5, kmax = 33000):
    """ Unimolecular hairpin closing rate, as a function of hairpin loop length. 

    References for default values: 
    Bonnet et al 1998, Kuznetsov et al 2008, Nayak et al 2012, Tsukanov et al 2013ab

    Args:
        hllen (flt): Hairpin loop length. The number of unpaired bases in a
            hairpin loop, or an arbitrary linker length in terms of an
            equivalently long chain of unpaired bases.
        ha (flt, optional): The rate for closing the first base-pair. 
            Defaults to 1_000_000 [/s].
        hb (flt, optional): Exponential factor to relate hllen to the
            probability of forming the initial contact. Defaults to -2.5.
        kmax (flt, optional): Fastest hairpin closing rate. Defaults to 33_000 [/s]. 
            (E.g. hairpins shorter than 4 nt cannot close any faster than this.)

    Returns:
        [flt]: The rate for closing the initial base-pair of a hairpin.
    """
    return min(ha * hllen**hb, kmax)


def binding_rate(length, before, after):
    """
    Rate constant formula for unimolecular binding of a domain of the given length.
    Could be zippering, hairpin closing, bubble closing, bulge formation, multiloop formation,
    depending on the flanking loops, which may be open or closed.
    """
    if not before.is_open and before.stems == 1 and before.bases == 0:
        if not after.is_open and after.stems == 1 and after.bases == 0:
            return 1e4  # bubble closing rate from Altan-Bonnet 2003
        # zippering from Wetmur&Davidson 1968, Gueron&Leroy 1995, Srinivas et al 2013, low end
        return (1e6) / length
    if not after.is_open and after.stems == 1 and after.bases == 0:
        # zippering from Wetmur&Davidson 1968, Gueron&Leroy 1995, Srinivas et al 2013, low end
        return (1e6) / length

    # bulge closing assumed to be similar to faster of two hairpin closings
    L = polymer_link_length(before, after)
    # hairpin closing adapted (simplified) from data in Bonnet et al 1998
    return polymer_link_rate(L)

# Diagram for 3-way branch migration, general case.
# Loops could be listed either 5'->3' or 3'->5, but they always go from
# invading domain to bound stem (non-inclusive).
#
#                before
#                _______  x (bound domain)
#               /       \____
# (invading) x |         ____
#               \_______/ x*
#
#                 after


def show_loops(before, after, message):
    """
    Debugging help: show (partial) loops returned by find_on_loop().  

    ! indicates a stem, | indicates open loop break.  
    """
    print('\n{}:'.format(message))
    stb = ''
    for step in before.parts:
        if step is None: # strand break?
            stb += " +"
        else:
            (dom, struct, loc) = step
            bound = '!' if struct is not None else ''
            stb += " {}{}".format(dom.name,bound)
    print("before: [{} ] is_open = {}, stems = {:d}, bases = {:d}".format(stb,
            before.is_open, before.stems, before.bases))
    sta = ''
    for step in after.parts:
        if step is None: # strand break?
            sta += " +"
        else:
            (dom, struct, loc) = step
            bound =  '!' if struct is not None else ''
            sta += " {}{}".format(dom.name,bound)
    print("after:  [{} ] is_open = {}, stems = {:d}, bases = {:d}".format(sta,
            after.is_open, after.stems, after.bases))

def branch_3way_remote_rate(length, before, after, debug = False):
    """
    Rate constant formula for 3-way branch migration, possibly with a remote toehold.

    Args: 
        length (int): the length of the domain
        before (:obj:Loop): the loop object on the typical 3-way branch migration initiation site.
        after (:obj:Loop): the other (remote) loop object.

    Returns:
        k1 = P(link) * k_init / l
    """
    # step = 0.1e-3  # 100us branch migration step time from Srinivas et al 2013 (not relevant)
    # k_init = k_uni exp(-dGsp / RT) with k_uni = 7.5e7, dGsp = 7.3 kcal/mol,
    # T = 273.15 + 25 K, R = 0.001987 kcal/mol/K

    # Experimentally derived time for a single step of 3-way branch migration.
    t_step = 1e-4  # sec
    k_step = 1/t_step

    if after._parts[0] is None: 
        # Initialtion penalty if there is *no* danling end on the displaced side.
        t_init = 3.0e-3 
        k_init = 1/t_init
    else :
        # Initialtion penalty if there is a danling end on the displaced side,
        # in which case it is just a regular displacement step.
        t_init = t_step
        k_init = k_step

    k1 = None # The probability of a successfull branch migration scales with 1/L
    k2 = None # The time spent on branch migration scales with 1/L^2

    if debug :
        show_loops(before, after, "before & after loops for 3-way branch migration")

    # "standard" 3-way bm initiation (plus "after" being closed)
    if not before.is_open and before.stems == 1 and before.bases == 0:
        k1 = k_init / length
        #k2 = k_init / length + k_step / length**2
    else :
        # consider a slowdown analogous to Genot et al 2011 (remote) supp info derivation
        # bulge closing assumed to be similar to faster of two hairpin closings
        L = polymer_link_length(before, after)

        # how much slower than our (admittedly slow) zippering rate is this?
        ratio = polymer_link_rate(L) / 1e6

        # we slow down initiation and thus success probability (note: ratio < 1/30)
        k1 = ratio * k_init / length
        #k2 = ratio * k_init / length + k_step / length**2

    return k1

def branch_4way_remote_rate(length, before, after, debug = False):
    """
    Rate constant formula for 4-way branch migration, possibly with a remote toehold
    """
    # rates recalculated from Nadine Dabby, PhD Thesis, 2013, based on
    # assuming last 6 bp dissociate faster than 4way bm step
    open_step = 107   # sec, = 1/k_first  (this is for open 4-way case only)
    # sec, = 1/k_bm     (this is used for initiating closed 4-way case;
    # consistent with Panyutin&Hsieh 1993)
    closed_step = 3.0

    # open_step = 200 # fudge !
    
    if debug:
        show_loops(before, after, "before & after loops for 4-way branch migration")

    if not before.is_open and not after.is_open:
        init = closed_step
        if before.bases == 0 and before.stems == 1 and after.bases == 0 and after.stems == 1:
            return 1.0 / init / length   # closed & ready-to-rock-and-roll 4 way initiation
    if before.is_open:
        init = open_step
        if after.bases == 0 and after.stems == 1:
            # we care about probability of taking this path, not actual time
            return 1.0 / init / length
    if after.is_open:
        init = open_step
        if before.bases == 0 and before.stems == 1:
            # we care about probability of taking this path, not actual time
            return 1.0 / init / length

    # consider a slowdown analogous to Genot et al 2011 (remote) supp info derivation
    # bulge closing assumed to be similar to faster of two hairpin closings
    L = polymer_link_length(before, after)
    # how much slower than our (admittedly slow) zippering rate is this?
    ratio = polymer_link_rate(L) / (1e6)
    # we slow down initiation and thus success probability (note: ratio < 1/30)
    return ratio / init / length

def bimolecular_binding_rate(length, k_nuc = 3e5):
    """ Rate constant formula for bimolecular association (binding).

    Hybridization is determined by well-aligned initial contacts, therefore 
    dependent linearly on length

    Literature: 
        Wetmur 1976 review,
        Srinivas et al. 2013 AEL model, 
        Zhang & Winfree 2009.

    Args:
        length (int): The length of a domain.
        k_nuc (flt, optional): The hybridization rate per nucleotide. 
            Defaults to 3*10^5.

    Returns:
        [float]: The rate constant for bimolecular binding.
    """
    return length * k_nuc

# Reaction functions
# ----------------------------------------------------------------------------

def bind11(reactant, max_helix=True, remote=None):
    """
    Returns a list of reaction pathways which can be produced by 1-1 binding
    reactions of the argument complex. The 1-1 binding reaction is the
    hybridization of two complementary unpaired domains within a single complex
    to produce a single unpseudoknotted product complex.

    Note: Remote is ineffective, but may be set for convencience
    """
    
    reactions = []
    structure = reactant.pair_table

    # We iterate through all the domains
    for (strand_index, strand) in enumerate(structure):
        for (domain_index, domain) in enumerate(strand):

            # The displacing domain must be free
            if structure[strand_index][domain_index] is not None :
                continue

            start_loc = (strand_index, domain_index)

            # search both directions around the loop for a bound domain that
            # has the same sequence (and therefore can be displaced)
            results = find_on_loop(reactant, start_loc, filter_bind11)

            if results:
                assert len(results) == \
                        len(find_on_loop(reactant, start_loc, filter_bind11, direction=-1))

            for e, (invader, xlinker, target, ylinker) in enumerate(results):
                if max_helix:
                    invader, xlinker, target, ylinker = zipper(
                            reactant, invader[0], xlinker, target[0], ylinker, filter_bind11)
                results[e] = list(map(Loop, [invader, xlinker, target, ylinker]))

            # build products
            for (loc1s, before, loc2s, after) in results:

                try :
                    (product,rotations) = do_bind11(reactant, loc1s.locs, loc2s.locs)
                except AssertionError:
                    continue

                try :
                    reaction = PepperReaction([reactant], [product], 'bind11')
                    reaction.meta = (loc1s, loc2s, before, after)
                    reaction.rotations = rotations
                    length = len(loc1s) # length of invading domain
                    reaction.const = binding_rate(length, before, after)
                except DSDDuplicationError as e :
                    reaction = e.existing

                reactions.append(reaction)

    # remove any duplicate reactions
    return sorted(list(set(reactions)))

def do_bind11(reactant, loc1s, loc2s, sanitycheck = False):
    """ Returns PepperComplex after the bind11 reaction. """
    struct = reactant.pair_table
    for loc1, loc2 in zip(loc1s, loc2s):
        assert struct[loc1[0]][loc1[1]] is None
        assert struct[loc2[0]][loc2[1]] is None
        struct[loc1[0]][loc1[1]] = loc2
        struct[loc2[0]][loc2[1]] = loc1
    newstr = pair_table_to_dot_bracket(struct)
    try:
        product = PepperComplex(reactant.sequence, newstr)
        product.pair_table = struct
        rotations = 0
    except DSDDuplicationError as e:
        product = e.existing
        rotations = e.rotations
    return (product, rotations)

def bind21(reactant1, reactant2, max_helix = True, remote=None, pkwarning=False):
    """
    Returns a list of reaction pathways which can be produced by 2-1 binding
    reactions of the argument complexes. The 2-1 binding reaction is the
    hybridization of two complementary unpaired domains, each in a different
    complex, to produce a single, unpseudoknotted product complex containing
    all of the strands contained in either of the original complexes.

    Note: remote is ineffective, but may be set for convencience
    """
    r1_doms = reactant1.available_domains
    r2_doms = reactant2.available_domains

    reactions = []

    # Iterate through all the free domains in reactant1
    for (dom1, strand_num1, dom_num1) in r1_doms:
        # For each, find any domains in reactant2 that could bind
        for (dom2, strand_num2, dom_num2) in r2_doms:
            # If it can pair, this is one possible reaction (this kind of
            # reaction cannot possibly produce a pseudoknotted structure)
            if (dom1.can_pair(dom2)):
                # combine the two complexes into one, but do not perform the association
                reactions.append(join_complexes_21(
                    reactant1, (strand_num1, dom_num1),
                    reactant2, (strand_num2, dom_num2)))

    if pkwarning: # check if potential pseudoknots are there:
        pk1_doms = reactant1.pk_domains
        pk2_doms = reactant2.pk_domains
        # Iterate through all the free domains in reactant1
        for (dom1, strand_num1, dom_num1) in r1_doms + pk1_doms:
            # For each, find any domains in reactant2 that could bind
            for (dom2, strand_num2, dom_num2) in r2_doms + pk2_doms:
                if (dom1, strand_num1, dom_num1) in r1_doms and \
                    (dom2, strand_num2, dom_num2) in r2_doms:
                    # Exclude the non-pseudoknotted interactions
                    continue
                if (dom1.can_pair(dom2)):
                    print("# WARNING: potential pk-interaction: {} and {}".format(
                        reactant1, reactant2))

   
    output = set()
    for complex, location1, location2 in reactions:
        def findloc(trip1, trip2):
            (dom1, struct1, loc1) = trip1 
            (dom2, struct2, loc2) = trip2
            return loc1 == location1 and loc2 == location2

        # build "before" and "after" loop structures via find_on_loop ...
        out = find_on_loop(complex, location1, findloc)

        [(loc1s, before, loc2s, after)] = out

        # zipper for max-helix semantics
        if max_helix :
            loc1s, before, loc2s, after = zipper(
                    complex, loc1s[0], before, loc2s[0], after, filter_bind11)

        [loc1s, before, loc2s, after] = list(map(Loop, [loc1s, before, loc2s, after]))
        (product,rotations) = do_bind11(complex, loc1s.locs, loc2s.locs)

        try :
            reaction = PepperReaction(sorted([reactant1, reactant2]), [product], 'bind21')
            reaction.const = bimolecular_binding_rate(len(loc1s))
        except DSDDuplicationError as e :
            reaction = e.existing

        output.add(reaction)

    return sorted(list(output))

def join_complexes_21(complex1, location1, complex2, location2):
    """
    Combines two complexes to form one complex, binding the domain in
    complex1 at location1 to the domain in complex2 at location2.

    Returns the new complex.
    """

    # make sure that this value is larger than any possible loc in the pair_table
    maxlen = complex1.size + complex2.size + 1

    if complex1.get_loop_index(location1) == 0 :
        seq1 = complex1.sequence
        ptb1 = complex1.pair_table
        loc1 = location1
        ptb1[loc1[0]][loc1[1]] = (maxlen,maxlen) # add an additional '(' 
    else :
        seen = False # don't break, otherwise you might lose the original strand ordering...
        for e, rot in enumerate(complex1.rotate(),1):
            tmp = complex1.rotate_location(location1, -e)
            if not seen and complex1.get_loop_index(tmp) == 0 :
                seq1 = complex1.sequence
                ptb1 = complex1.pair_table
                loc1 = tmp
                ptb1[loc1[0]][loc1[1]] = (maxlen,maxlen) # add an additional '(' 
                seen = True

    if complex2.get_loop_index(location2) == 0 :
        seq2 = complex2.sequence
        ptb2 = complex2.pair_table
        loc2 = location2
        ptb2[loc2[0]][loc2[1]] = (-1,-1) # add an additional ')' 
    else :
        seen = False # don't break, otherwise you might lose the original strand ordering...
        for e, rot in enumerate(complex2.rotate(),1):
            tmp = complex2.rotate_location(location2, -e)
            if not seen and complex2.get_loop_index(tmp) == 0 :
                seq2 = complex2.sequence
                ptb2 = complex2.pair_table
                loc2 = tmp
                ptb2[loc2[0]][loc2[1]] = (-1,-1) # add an additional ')' 
                seen = True

    # build the new sequence and structure *including* the new pair
    newseq = seq1 + ['+'] + seq2
    newstr = pair_table_to_dot_bracket(ptb1) + ['+'] + pair_table_to_dot_bracket(ptb2)

    # update loc2 from the new structure
    combined = make_pair_table(newstr)
    loc2 = combined[loc1[0]][loc1[1]]

    # remove the new pair again ...
    combined[loc1[0]][loc1[1]] = None
    combined[loc2[0]][loc2[1]] = None

    # update the structure to the unpaired (disconnected) version.
    newstr = pair_table_to_dot_bracket(combined)

    try :
        new_complex = PepperComplex(newseq, newstr)
    except DSDDuplicationError as e:
        new_complex = e.existing
        # strands may be re-ordered in new complex, so we need to back
        # out where the new strands ended up
        loc1 = new_complex.rotate_location(loc1, e.rotations)
        loc2 = new_complex.rotate_location(loc2, e.rotations)

    if loc1 > loc2:
        (loc1, loc2) = (loc2,loc1)

    return new_complex, loc1, loc2

def open(reactant, max_helix = True, release_11 = 6, release_1N = 6, dG_bp = -1.7):
    """ Returns a list of open reactions.

    Args:
        reactant (PepperComplex): The reactant complex
        max_helix (bool, optional): Use max-helix notion. Defaults to True.
        release_11 (int, optional): Threshold length for a open11 reaction. Defaults to 6.
        release_1N (int, optional): Threshold length for a open12 reaction. Defaults to 6.

    Returns:
        [PepperReactions]

    """

    # remember the larger release cutoff; don't enumerate any reactions
    # for helices longer than this
    max_release = max(release_11, release_1N) if release_11 and release_1N else 0

    # Disabled breathing option for now. 
    # It is effectively the same as not using max-helix.
    if False and max_release:
        end_release = (max_release + 1) / 2
    else :
        end_release = 0

    reactions = []
    structure = reactant.pair_table

    def triple(loc):
        return (reactant.get_domain(loc), reactant.get_structure(loc), loc)

    # for no-max-helix mode:
    if not max_helix:
        # We iterate through all the domains
        for (strand_index, strand) in enumerate(structure):
            for (domain_index, domain) in enumerate(strand):

                # The bound domain must be... bound
                if reactant.get_structure((strand_index, domain_index)) is None :
                    continue

                bound_loc = (strand_index, domain_index)
                bound_domain = reactant.get_domain(bound_loc)

                # Assume we want to correct for the actual number of stacks,
                # not just use the length of a helix, then this code can be helpful...
                #
                # thispair = reactant.get_structure((strand_index, domain_index))
                # try:
                #     prevpair = reactant.get_structure((strand_index, domain_index-1))
                #     nextpair = reactant.get_structure((strand_index, domain_index+1))
                #     if prevpair and nextpair and \
                #             prevpair[0] == thispair[0] and nextpair[0] == thispair[0] and \
                #             prevpair[1] == thispair[1]+1 and nextpair[1] == thispair[1]-1:
                #                 continue
                # except IndexError:
                #     pass

                (release_reactant, rotations) = do_single_open(reactant, bound_loc)
                product_set = release_reactant.split()
                meta = (Loop([triple(bound_loc)]),
                        Loop([triple(reactant.get_structure(bound_loc))]), None, None)
                reactions.append((product_set, rotations, len(bound_domain),
                    meta))

    else: # for max-helix mode:
        # We loop through all stands, domains
        for (strand_index, strand) in enumerate(structure):
            for (domain_index, domain) in enumerate(strand):
                # If the domain is unpaired, skip it
                if (structure[strand_index][domain_index] is None):
                    continue

                # A: Strand/domain position on "top" strand - CG 5/21
                helix_startA = [strand_index, domain_index]

                # B: Strand/domain position on "bottom" strand - CG 5/21
                helix_startB = list(structure[strand_index][domain_index])

                # If the domain is bound to an earlier domain, then we have
                # already considered it, so skip it
                if (helix_startB < helix_startA):
                    continue

                helix_endA = helix_startA[:]
                helix_endB = helix_startB[:]

                bound_loc = (strand_index, domain_index)
                bound_domain = reactant.get_domain(bound_loc)
                helix_length = len(bound_domain)

                if max_release and helix_length > max_release:
                    continue

                ext_fw = False
                ext_bw = False

                # Now iterate through the whole helix to find the other end of
                # this one (The helix ends at the first strand break from
                # either direction)
                while True:
                    # Strands run in opposite directions, so A must be incremented
                    # and B decremented in order that both pointers move "right"
                    # along the helix- CG 5/21
                    helix_endA[1] += 1
                    helix_endB[1] -= 1

                    # If one of the strands has broken, the helix has ended
                    if helix_endA[1] >= reactant.strand_length(helix_endA[0]) :
                        break
                    elif helix_endB[1] < 0 :
                        break

                    # If these domains aren't bound to each other, the helix has ended
                    if (tuple(helix_endA) != structure[helix_endB[0]][helix_endB[1]]):
                        break

                    # Add the current domain to the current helix
                    temp_bl = (helix_endA[0], helix_endA[1])
                    helix_length += len(reactant.get_domain(temp_bl))

                    ext_fw = True

                # We must also iterate in the other direction
                while True:
                    helix_startA[1] -= 1
                    helix_startB[1] += 1

                    # If one of the strands has broken, the helix has ended
                    if helix_startA[1] < 0 :
                        break
                    elif helix_startB[1] >= reactant.strand_length(helix_startB[0]):
                        break

                    # If these domains aren't bound to each other, the helix has ended
                    if tuple(helix_startA) != structure[helix_startB[0]][helix_startB[1]] :
                        break

                    # Add the current domain to the current helix
                    temp_bl = (helix_startA[0], helix_startA[1])
                    helix_length += len(reactant.get_domain(temp_bl))

                    ext_bw = True

                # Move start location back to the first domain in the helix
                helix_startA[1] += 1
                helix_startB[1] -= 1

                if ext_fw and ext_bw :
                    # we only want to allow moves that start at helix ends!
                    continue

                # If the helix is short enough, we have a reaction
                if not max_release or helix_length <= max_release:

                    new_struct = reactant.pair_table

                    # Delete all the pairs in the released helix
                    for dom in range(helix_startA[1], helix_endA[1]):
                        bound_loc = reactant.get_structure((helix_startA[0], dom))
                        new_struct[helix_startA[0]][dom] = None
                        new_struct[bound_loc[0]][bound_loc[1]] = None

                    newstr = pair_table_to_dot_bracket(new_struct)

                    try:
                        release_reactant = PepperComplex(reactant.sequence, newstr)
                        rotations = 0
                    except DSDDuplicationError as e:
                        release_reactant = e.existing
                        rotations = e.rotations

                    product_set = release_reactant.split()
                    meta = None
                    reactions.append((product_set, rotations, helix_length, meta))

                elif len(bound_domain) <= end_release: # bound-domain is initial domain
                    new_struct = reactant.pair_table

                    # Delete all the pairs in the released helix
                    breathing = list(range(helix_startA[1], helix_endA[1]))
                    if ext_bw:
                        breathing = reversed(breathing)

                    bl = 0
                    for dom in breathing:
                        bound_loc = reactant.get_structure((helix_startA[0], dom))
                        dl = len(reactant.get_domain(bound_loc))
                        if bl + dl > end_release :
                            break
                        bl += dl
                        new_struct[helix_startA[0]][dom] = None
                        new_struct[bound_loc[0]][bound_loc[1]] = None

                    newstr = pair_table_to_dot_bracket(new_struct)

                    try:
                        release_reactant = PepperComplex(reactant.sequence, newstr)
                        rotations = 0
                    except DSDDuplicationError as e:
                        release_reactant = e.existing
                        rotations = e.rotations

                    product_set = release_reactant.split()
                    meta = None
                    reactions.append((product_set, rotations, bl, meta))

    output = []
    for product_set, rotations, length, meta in reactions:
        try :
            reaction = PepperReaction([reactant], sorted(product_set), 'open')
            reaction.rotations = rotations
            reaction.meta = meta
            reaction.const = opening_rate(length, dG_bp = dG_bp)
        except DSDDuplicationError as e :
            reaction = e.existing

        # discard reactions where the release cutoff is greater than the threshold
        if release_11 and len(reaction.products) == 1 and length > release_11:
            continue
        elif release_1N and len(reaction.products) > 1 and length > release_1N:
            continue

        output.append(reaction)

    return sorted(list(set(output)))

def do_single_open(reactant, loc):
    struct = reactant.pair_table
    loc1 = loc
    loc2 = struct[loc1[0]][loc1[1]]
    assert struct[loc2[0]][loc2[1]] == loc1
    struct[loc1[0]][loc1[1]] = None
    struct[loc2[0]][loc2[1]] = None
    newstr = pair_table_to_dot_bracket(struct)
    try:
        product = PepperComplex(reactant.sequence, newstr)
        product.pair_table = struct
        rotations = 0
    except DSDDuplicationError as e:
        product = e.existing
        rotations = e.rotations
    return (product, rotations)

def branch_3way(reactant, max_helix = True, remote=True):
    """
    Returns a list of reaction pathways that can be created through one
    iteration of a 3 way branch migration reaction (more than one molecule may
    be produced by a reaction because branch migration can liberate strands and
    complexes).
    """

    reactions = []

    reactions = []
    structure = reactant.pair_table

    # We iterate through all the domains
    for (strand_index, strand) in enumerate(structure):
        for (domain_index, domain) in enumerate(strand):

            # The displacing domain must be free
            if (structure[strand_index][domain_index] is not None):
                continue

            # search 5'->3' and 3'->5' directions around the loop for a bound
            # domain that is complementary (and therefore can be displaced)
            start_loc = (strand_index, domain_index)

            # build products
            fwresults = find_on_loop(reactant, start_loc, filter_3way, direction=1)
            bwresults = find_on_loop(reactant, start_loc, filter_3way, direction=-1)

            results = []
            for (invader, xlinker, target, ylinker) in fwresults:
                if max_helix:
                    invader, xlinker, target, ylinker = zipper(
                            reactant, invader[0], xlinker, target[0], ylinker, filter_3way)
                ylinker += invader
                results.append(list(map(Loop, [invader[::-1], xlinker, target[::-1], ylinker])))

            for (invader, xlinker, target, ylinker) in bwresults:
                if max_helix:
                    invader, xlinker, target, ylinker = zipper(
                            reactant, invader[0], xlinker, target[0], ylinker, filter_3way)
                ylinker += invader[::-1]
                results.append(list(map(Loop, [invader, xlinker, target, ylinker])))

            for (displacing, before, bound, after) in results:
                (products, rotations) = do_3way_migration(reactant,
                        list(displacing.locs), list(bound.locs))

                try :
                    reaction = PepperReaction([reactant], products, 'branch-3way')
                    reaction.meta = (displacing, bound, before, after)
                    reaction.rotations = rotations
                    reaction.const = branch_3way_remote_rate(len(displacing), before, after)
                except DSDDuplicationError as e :
                    reaction = e.existing

                # skip remote toehold reactions if directed
                if not remote :
                    if not (not before.is_open and before.stems == 1 and before.bases == 0):
                        # print "Rejecting... " + reaction.kernel_string
                        # import pdb; pdb.set_trace()
                        continue

                # calculate reaction constant
                reactions.append(reaction)

    # Remove any duplicate reactions
    return sorted(list(set(reactions)))

def do_3way_migration(reactant, displacing_locs, bound_locs):
    """
    Each location in displacing_locs will end up bound to the corresponding
    location in bound_locs. The stuff bound to bound_locs will end up un-bound
    """

    def update_structure(struct, displacing_loc, new_bound_loc):
        """
        displacing_loc will be bound to new_bound_loc; whatever new_bound_loc
        was bound to will be unbound.
        """
        displaced_loc = struct[new_bound_loc[0]][new_bound_loc[1]]

        assert struct[displacing_loc[0]][displacing_loc[1]] is None
        assert struct[new_bound_loc[0]][new_bound_loc[1]] is not None
        assert struct[displaced_loc[0]][displaced_loc[1]] is not None
        assert struct[displaced_loc[0]][displaced_loc[1]] == new_bound_loc
    
        struct[displacing_loc[0]][displacing_loc[1]] = new_bound_loc
        struct[new_bound_loc[0]][new_bound_loc[1]] = displacing_loc
        struct[displaced_loc[0]][displaced_loc[1]] = None

    struct = reactant.pair_table

    for displacing_loc, new_bound_loc in zip(displacing_locs, bound_locs):
        update_structure(struct, displacing_loc, new_bound_loc)

    newstr = pair_table_to_dot_bracket(struct)

    try:
        product = PepperComplex(reactant.sequence, newstr)
        product.pair_table = struct
        rotations = 0
    except DSDDuplicationError as e:
        product = e.existing
        rotations = e.rotations
  
    return (product.split(), rotations)

def branch_4way(reactant, max_helix = False, remote=True):
    """
    Returns a list of complex sets that can be created through one iteration of
    a 4 way branch migration reaction (each set consists of the molecules that
    result from the iteration; more than one molecule may result because branch
    migration can liberate strands and complexes).
    """

    reactions = []
    structure = reactant.pair_table

    # We loop through all domains
    for (strand_index, strand) in enumerate(structure):
        for (domain_index, domain) in enumerate(strand):

            # Unbound domains can't participate in branch migration
            if (structure[strand_index][domain_index] is None):
                continue

            start_loc = (strand_index, domain_index)

            # searches only 5'->3' direction around the loop for a bound domain that
            # has the same sequence (and therefore can be displaced)
            #
            #   z  _~_  z* (displacing)
            #  ___/   \___>
            #
            # <___     ___
            #     \_ _/
            #   z*  ~   z
            #

            # build products
            results = find_on_loop(reactant, start_loc, filter_4way)

            for e, (invader, xlinker, target, ylinker) in enumerate(results):
                if max_helix:
                    invader, _, target, _ = zipper(
                            reactant, invader[0], None, target[0], None, filter_4way)
                results[e] = list(map(Loop, [invader, xlinker, target, ylinker]))

            for (displacing, before, displaced, after) in results:
                (products,rotations) = do_4way_migration(reactant, displacing.locs,
                            (structure[dl[0]][dl[1]] for dl in displacing.locs),
                            (structure[bl[0]][bl[1]] for bl in displaced.locs), 
                            displaced.locs)

                try :
                    reaction = PepperReaction([reactant], products, 'branch-4way')
                    reaction.meta = (displacing, displaced, before, after)
                    reaction.rotations = rotations
                    reaction.const = branch_4way_remote_rate(
                            len(displacing), before, after)
                except DSDDuplicationError as e :
                    reaction = e.existing

                # skip remote toehold reactions
                if not remote:
                    # NOTE: both sides need to be remote!
                    if not ((not after.is_open and after.stems == 1 and after.bases == 0) or
                            (not before.is_open and before.stems == 1 and before.bases == 0)):
                        continue

                reactions.append(reaction)

    # remove any duplicate reactions
    return sorted(list(set(reactions)))

def do_4way_migration(reactant, loc1s, loc2s, loc3s, loc4s):
    """
    Perform a sequence of max_helix 4-way branch migration reactions.
    """
    def update_structure(struct, loc1, loc2, loc3, loc4):
        """
        Performs a 4 way branch migration on a copy of reactant, with loc1 as the
        displacing domain, loc2 as the domain displaced from loc1, loc3 as the
        template domain, and loc4 as the domain displaced from loc3. Returns the
        set of complexes produced by this reaction (may be one or more complexes).
    
        loc1:loc2, loc3:loc4 -> loc1:loc3, loc2:loc4
        """
        assert None not in (loc1, loc2, loc3, loc4)
        assert struct[loc1[0]][loc1[1]] == loc2
        assert struct[loc3[0]][loc3[1]] == loc4
        assert struct[loc2[0]][loc2[1]] == loc1
        assert struct[loc4[0]][loc4[1]] == loc3

        struct[loc1[0]][loc1[1]] = loc3
        struct[loc3[0]][loc3[1]] = loc1
        struct[loc2[0]][loc2[1]] = loc4
        struct[loc4[0]][loc4[1]] = loc2

    struct = reactant.pair_table
    for loc1, loc2, loc3, loc4 in zip(loc1s, loc2s, loc3s, loc4s):
        update_structure(struct, loc1, loc2, loc3, loc4)
    newstr = pair_table_to_dot_bracket(struct)
    try:
        product = PepperComplex(reactant.sequence, newstr)
        product.pair_table = struct
        rotations = 0
    except DSDDuplicationError as e:
        product = e.existing
        rotations = e.rotations
    return (product.split(), rotations)

# Filter functions for find_on_loop()
def filter_bind11(trip1, trip2):
    (dom1, struct1, loc1) = trip1
    (dom2, struct2, loc2) = trip2
    return struct1 is None and struct2 is None and dom2.can_pair(dom1)

def filter_3way(trip1, trip2):
    (dom1, struct1, loc1) = trip1 
    (dom2, struct2, loc2) = trip2 
    return (struct1 is None) and (struct2 is not None) and dom1.can_pair(dom2)

def filter_4way(trip1, trip2):
    (dom1, struct1, loc1) = trip1
    (dom2, struct2, loc2) = trip2
    return struct1 is not None and struct2 is not None and dom1 == dom2

def find_on_loop(reactant, start_loc, pattern, direction=1):
    r"""Find a reaction pattern within a loop.
    
    Starts at a particular locus and searches for every possible pattern that
    preseves secondary structure (matches within the "loop").  Where a loop
    involves stems, only one of the complementary domains will be listed in the
    array of tuples, specifically, the "first" one in the search direction.
    Thus, a multiloop with n unpaired domains and m stems will result, for
    closed loops, in `len(before+after) == n+m-2`, as the match location and
    `start_loc` are omitted.

    Note 1: `before` and `after` refer to the partial loops between `start_loc`
    and each of the results, _in the `direction`_ of the search. For example:

       A
          ____
         /    \ 
     x  |     |  x*
        |
         \____> 3'

            B

    If `start_loc` pointed to `x` and `direction` is +1, then `before` would
    be `A` and `after` would be `B`. If instead `direction` is -1, then
    `before` is `B` and `after` is `A`.  
    Note 2: If the domain passed to `start_loc` is a duplex, the results may
    be unexpected:

           ___  x  ___
    5' ___/   \___/   \___
    3' ___  A  ___  B  ___)
          \___/   \___/
                x*

    Notice that the duplex x() participates in two internal loops (A and B).
    By convention, the internal loop considered is the _internal loop which
    encloses this domain_. That means if you pass domain x and +1, you'll get
    loop A, whereas if you pass x and -1, you'll get loop B. This is in an
    attempt to be consistent with the case where you pass an unpaired domain
    (and therefore the internal loop searched is the one which encloses the
    unpaired domain).
  
    Args:
        reactant (:obj:) = The reactant complex object.
        start_loc ((int,int)) = A tuple pointing to the respective strand and
            domain indices for reactant.pair_table
        pattern (function) = A pointer to a pattern match function which
            determines if there exists a valid binding partner for the start locus.
            The function takes two arguments containing (domain, paired, locus)
            and returns True or False.
        directions (optional: -1,+1): Searching for binding partners in 
            5'->3' (+1) or 3'->5' (-1) direction.

    Returns: 
        TODO
    """

    results = []
    loop = []
    def triple(loc):
        return (reactant.get_domain(loc), reactant.get_structure(loc), loc)

    # We now follow the external loop from the starting pair
    # searching for a bound domain to displace
    bound_loc = start_loc

    # Avoid getting stuck inside an internal loop enclosed by this domain,
    # if the starting domain is a duplex.
    #
    #   1      2
    #  ___________
    #  ____  _____
    #   1*  /  2*
    #
    #  If we start at domain 1, going in the - direction, then
    #  immediately continue to the next domain, we'll go to 1*
    if reactant.get_structure(bound_loc) is not None:
        bound_loc = reactant.get_structure(bound_loc)
    
    # Follow the external loop to the end
    while True:
        # move to the next domain in the indicated direction
        # (+1 = 5' -> 3', -1 = 3' -> 5')
        bound_loc = (bound_loc[0], bound_loc[1] + direction)

        # if we've reached the end of the strand (5')
        if (bound_loc[1] == -1):
            # Continue to next strand
            bound_loc = (wrap(bound_loc[0] - 1, reactant.size),)
            bound_loc = (bound_loc[0], reactant.strand_length(bound_loc[0]))
            loop.append(None)
            continue

        # if we've reached the end of the strand (3')
        if (bound_loc[1] == reactant.strand_length(bound_loc[0])):
            # Continue to next strand
            bound_loc = (wrap(bound_loc[0] + 1, reactant.size), -1)
            loop.append(None)
            continue

        if bound_loc == start_loc:
            # We've returned to the original location of the displacing domain
            break

        # try to match the pattern function
        if pattern(triple(start_loc), triple(bound_loc)):
            # append the location
            results.append((bound_loc, len(loop)))

        # store unpaired domains and "first" domain of each stem
        loop.append(triple(bound_loc))
 
        # if the domain at bound_loc is unbound
        if reactant.get_structure(bound_loc) is None:
            # look to the next domain
            continue
        else :
            # follow the structure to stay on the same loop
            bound_loc = reactant.get_structure(bound_loc)

    return [([triple(start_loc)],   # invading domain
             loop[:i],              # first linker 
             [triple(bound_loc)],   # target domain
             loop[i+1:]) for (bound_loc, i) in results]

def zipper(reactant, start_trp, before, bound_trp, after, pattern):
    r"""Max-helix mode zipping to extend a given move type.

    Takes a result from `find_on_loop` and finds as many adjacent domains as
    possible such that the `pattern` function still returns True.

    For example, if `start_loc` was b1 and `bound_loc` was b1*, and the pattern
    function specified that the domain at `start_loc` must be complementary to
    the domain at `bound_loc`, then the function would return [b1,b2] as
    start_locs and [b1*, b2*] as bound_locs

            b1* b2*
            ______
         __/      \__>
        <__        __
           \______/
            b1  b2

    Note that 4-way branch migration max-helix never extends into (changes) the 
    before / after loops, so it is safe to pass None instead.

    Returns:
        start_locs (extended list of triples)
        before (reduced list of triples)
        bound_locs (extended list of triples)
        after (reduced list of triples)
    """

    def triple(loc):
        return (reactant.get_domain(loc), reactant.get_structure(loc), loc)

    assert pattern(start_trp,bound_trp)

    def extend_match(start_loc, bound_loc, sdir, bdir):
        """
        Move upstream or downstream from start locus and bound locus and check
        if pattern condition still applies.

        Accesses reactant, start_loc, bound_loc, pattern, 
        
        """
        (sstrand, sdomain) = start_loc
        (bstrand, bdomain) = bound_loc

        start_pair = reactant.get_structure(start_loc)
        bound_pair = reactant.get_structure(bound_loc)

        sloc = (sstrand, sdomain + sdir)
        bloc = (bstrand, bdomain + bdir)

        try:
            spair = reactant.get_structure(sloc)
            bpair = reactant.get_structure(bloc)
        except IndexError:
            # sdomain is no longer on sstrand or 
            # bdomain is no longer on bstrand
            return

        if (start_pair is None) != (spair is None): return
        if (bound_pair is None) != (bpair is None): return
        if sloc == start_pair or bloc == bound_pair: return

            # ddomain hasn't passed bound_loc
        if ((#TODO: python 2->3 code.. there must be a nicer way
            #(cmp(sloc, bound_loc) == cmp(start_loc, bound_loc))
            (sloc > bound_loc) - (sloc < bound_loc) == \
            (start_loc > bound_loc) - (start_loc < bound_loc)) and
            # and bdomain hasn't passed start_loc
            (
            #cmp(bloc, start_loc) == cmp(bound_loc, start_loc)
            (bloc > start_loc) - (bloc < start_loc) == \
            (bound_loc > start_loc) - (bound_loc < start_loc)) and
            # if paired, then also the paired domain must be adjacent
            ((start_pair is None) and (spair is None) or
                start_pair[0] == spair[0] and         # same strand!
                start_pair[1] == spair[1] + sdir) and # adjacent!

            # if paired, then also the paired domain must be adjacent
            ((bound_pair is None) and (bpair is None) or
                bound_pair[0] == bpair[0] and         # same strand!
                bound_pair[1] == bpair[1] + bdir) and # adjacent!

            # and pattern condition still applies
            pattern(triple(sloc), triple(bloc))):

            # add new positions to list
            if sdir == 1:
                start_locs.append(triple(sloc))
                if before and before[0] == triple(sloc):
                    before.pop(0)
            else :
                start_locs[:0] = [triple(sloc)]
                if after and after[-1] == triple(sloc):
                    after.pop(-1)

            if bdir == 1:
                bound_locs[:0] = [triple(bloc)]
                if after and after[0] == triple(bloc):
                    after.pop(0)
            else :
                bound_locs.append(triple(bloc))
                if before and before[-1] == triple(bloc):
                    before.pop(-1)

            extend_match(sloc, bloc, sdir, bdir)

        return

    start_locs = [start_trp]
    bound_locs = [bound_trp]

    if filter_4way(start_trp,bound_trp):
        extend_match(start_trp[2], bound_trp[2], sdir = 1, bdir = 1)
        bound_locs = bound_locs[::-1]
    else :
        extend_match(start_trp[2], bound_trp[2], sdir = 1, bdir = -1)
        extend_match(start_trp[2], bound_trp[2], sdir = -1, bdir = 1)

    return start_locs, before, bound_locs, after




#
#  __init__.py
#  EnumeratorProject
#
__version__ = "v0.9"

import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())

from peppercornenumerator.enumerator import Enumerator, PolymerizationError
from peppercornenumerator.enumerator import enumerate_pil, enumerate_ssw
from peppercornenumerator.condense import CondensationError, PepperCondensation

# Deprecated since v0.6
from peppercornenumerator.condense import ReactionGraph




#
#  peppercornenumerator/condense.py
#  EnumeratorProject
#
from __future__ import absolute_import, print_function, division

import logging
log = logging.getLogger(__name__)

import operator
import collections
import itertools as it
import numpy as np
from functools import reduce

from peppercornenumerator.objects import PepperComplex
from peppercornenumerator.objects import PepperReaction
from peppercornenumerator.objects import PepperMacrostate
from peppercornenumerator.objects import DSDDuplicationError

class CondensationError(Exception):
    pass

class PepperCondensation(object):
    def __init__(self, enumerator):
        self._enumerator = enumerator

        # Condensed graph components
        self._SCCs = None
        self._set_to_fate = dict()
        self._complex_fates  = dict()
        self._condensed_reactions = None

        # Helper dictionaries
        self._scc_containing = None
        self._reactions_consuming = None

        # Condensed reaction rate calculations
        self._exit_probabilities = dict()
        self._stationary_distributions = dict()
        self._decay_probabilities = collections.defaultdict(float)
        self._reaction_decay_probabilities = collections.defaultdict(float)

    @property
    def enumerator(self):
        return self._enumerator

    @property
    def complexes(self):
        return self._enumerator.complexes

    @property
    def resting_sets(self):
        print("# Deprecated function: PepperCondensation.resting_sets. Replace with PepperCondensation.resting_macrostates")
        return self.resting_macrostates

    @property
    def resting_macrostates(self):
        if not self._set_to_fate:
            raise CondensationError('need to call condense first')
        return list(self._set_to_fate.values())

    @property
    def resting_set_representatives(self):
        if not self._set_to_fate:
            raise CondensationError('need to call condense first')
        return [x.canonical for x in list(self._set_to_fate.values())]

    @property
    def set_to_fate(self):
        if not self._set_to_fate:
            raise CondensationError('need to call condense first')
        return self._set_to_fate

    @property
    def cplx_to_fate(self):
        if not self._complex_fates:
            raise CondensationError('need to call condense first')
        return self._complex_fates

    @property
    def reactions(self):
        return self.condensed_reactions + self.detailed_reactions

    @property
    def detailed_reactions(self):
        return self._enumerator.reactions

    @property
    def condensed_reactions(self):
        if self._condensed_reactions is None:
            self.condense()
        return list(self._condensed_reactions)

    @property
    def k_fast(self):
        return self._enumerator.k_fast

    def is_fast(self, rxn):
        if rxn.arity[0] != 1:
            return False
        if self._enumerator.local_elevation:
            from peppercornenumerator.enumerator import local_elevation_rate
            k = local_elevation_rate(rxn)
        else :
            k = rxn.const
        return k >= self.k_fast

    def reactions_consuming(self, cplx):
        if self._reactions_consuming is None:
            self._reactions_consuming = get_reactions_consuming(self.complexes, self.reactions)
        return self._reactions_consuming[cplx]

    @property
    def SCCs(self):
        """
        Return (compute) a list of SCCs
        """
        if self._SCCs is None:
            if not self._reactions_consuming:
                self._reactions_consuming = get_reactions_consuming(
                        self.complexes, self.detailed_reactions)
            self._SCCs = tarjans(self.complexes, self.reactions, 
                    self._reactions_consuming, self.is_fast)
        return self._SCCs

    def SCC_containing(self, cplx):
        """ Map each complex to the SCC which contains the complex. 
        Each complex should be in 1 SCC.
        """
        if self._scc_containing is None:
            self._scc_containing = {c: scc for scc in self.SCCs for c in scc}
        return self._scc_containing[cplx]

    def calculate_reaction_decay_probabilities(self, rxn, fate, combinations=None):
        """
        Calculate the decay probability of a reaction to a particular fate.

        Each combination (`fates`) that sums to `fate` constitutes a possible
        way this reaction can decay to `fate`, and therefore contributes to 
        P(r -> fate). This contribution is the joint probability that each
        product `d` of r decays to the corresponding fate `f` in `fates`.
        """
        if combinations is None:
            combinations = cartesian_product(list(map(self.get_fates, rxn.products)))

        # NOTE: introdcued as a consequence of a bug discovered in the cooperative
        # binding case. Before you calculate (add to) reaction decay
        # probabilities, make sure that the dictionary is empty.
        self._reaction_decay_probabilities = collections.defaultdict(float)

        for fates in combinations :
            if tuple_sum_sort(fates) != fate:
                continue

            # NOTE: this code is equivalent to the code below, but seems to be
            # a little less accurate!?

            #p = 1
            #for (d, f) in zip(rxn.products, fates):
            #    p *= self._decay_probabilities[(d, f)]
            #self._reaction_decay_probabilities[(rxn, fate)] += p

            self._reaction_decay_probabilities[(rxn, fate)] += times(
                    self._decay_probabilities[(d, f)] for (d, f) in zip(rxn.products, fates))

    def get_fates(self, cplx):
        if cplx not in self._complex_fates :
            self.compute_fates(self.SCC_containing(cplx))
        return self._complex_fates[cplx]

    def compute_fates(self, scc):
        """
        Processes a single SCC neighborhood, generating resting set multisets
        for each complex, and storing the mappings in the outer-scope
        `complex_fates` dict

        """
        complex_fates = self._complex_fates

        if scc[0] in complex_fates:
            # Dirty check to see if the scc has been processed before.
            return 

        # Convert to a set for fast lookup
        scc_set = frozenset(scc)

        outgoing_reactions = []
        for c in scc:
            for r in self.reactions_consuming(c):
                if self.is_fast(r) and is_outgoing(r, scc_set) :
                    outgoing_reactions.append(r)

        # If this SCC is a resting set:
        if len(outgoing_reactions) == 0 :
            # build new resting set
            try :
                resting_set = PepperMacrostate(scc)
            except DSDDuplicationError as e:
                resting_set = e.existing

            self._set_to_fate[scc_set] = resting_set

            # calculate stationary distribution
            self._stationary_distributions[resting_set] = self.get_stationary_distribution(scc)

            # assign fate to each complex in the SCC
            fate = (resting_set,) # needs to be iterable..
            fates = frozenset([fate])

            for c in scc:
                if c in complex_fates:
                    raise CondensationError('complex should not be assigned yet')
                complex_fates[c] = SetOfFates(fates)

                # all complexes in this SCC decay to this SCC with probability 1,
                # e.g. P(c -> SCC) = 1 for all c in this SCC
                self._decay_probabilities[(c, fate)] = 1.0

        else :
            # Compute all possible combinations of the fates of each product of r
            reaction_fate_combinations = [cartesian_product(
                list(map(self.get_fates, rxn.products))) for rxn in outgoing_reactions]

            # Compute the fates of each of the outgoing reactions by summing
            # each element above. Returns a list, with each element
            # corresponding to the frozenset of fates for one reaction.
            reaction_fates = [sorted(map(tuple_sum_sort, combination))
                              for combination in reaction_fate_combinations]
            
            # note that these two are equivalent; the intermediate
            # reaction_fate_combinations is only calculated for rates
            assert reaction_fates == [sorted(cartesian_sum(
                list(map(self.get_fates, rxn.products)))) for rxn in outgoing_reactions]

            # calculate the exit probabilities
            self._exit_probabilities[scc_set] = self.get_exit_probabilities(scc)

            # calculate the probability of each fate:
            # for each outgoing reaction, `r`
            for (i, r) in enumerate(outgoing_reactions):
                # for each possible `fate` of that reaction `r`
                for fate in reaction_fates[i]:
                    # NOTE: this is weird, first you calculate the reaction decay
                    # probabilities from decay probabilities, then calculate decay
                    # probabilities from reaction decay probabilities?
                    fate = tuple(sorted(fate))
                    # calculate the probability that the products of `r`
                    # decay to `fate`, e.g. P(r -> fate)

                    # iterate over all combinations of the fates of `r`
                    # that sum to equal `fate`
                    combinations = reaction_fate_combinations[i]
                    self.calculate_reaction_decay_probabilities(r, fate, combinations)

                    # the decay probability will be different for different
                    # complexes in the SCC
                    for c in scc:
                        # P(x decays to F) = P(SCC exits via r | SCC was entered via c) 
                        #   * P(r decays to F)
                        self._decay_probabilities[(c, fate)] += self._exit_probabilities[scc_set][
                                (c, r)] * self._reaction_decay_probabilities[(r, fate)]

            # The set of fates for the complexes in this SCC is the union of
            # the fates for all outgoing reactions.

            # Note that frozenset().union(*X) === X[0] U X[1] U X[2] ...
            # where X[i] is a frozenset and a U b represents the union of
            # frozensets a and b
            set_of_fates = frozenset().union(*reaction_fates)

            for c in scc:
                if c in complex_fates:
                    raise CondensationError('complex should not be assigned yet')
                complex_fates[c] = SetOfFates(set_of_fates)

    # Main function!
    def condense(self):
        """
        Reaction condensation.
        """
        if self._condensed_reactions is not None:
            raise CondensationError('condensation called twice')
        else :
            self._condensed_reactions = set()

        enum = self.enumerator

        for scc in self.SCCs:
            self.compute_fates(scc)

        for reaction in enum.reactions :
            # NOTE: Fast reactions have been handled by the fates
            if self.is_fast(reaction):
                continue

            # Filter reactions with no products/reactants
            if len(reaction.reactants) == 0 and len(reaction.products) == 0:
                raise Exception('before we continue, see why that happens...')
                #continue

            # Get the corresponding fates (resting sets)
            reactant_fates = list(map(self.get_fates, reaction.reactants))
            product_fates = list(map(self.get_fates, reaction.products))

            # Get all combinations of reactant and product fates,
            new_reactant_combinations = cartesian_sum(reactant_fates)
            new_product_combinations = cartesian_sum(product_fates)

            # Make sure each reactant is a resting set: 
            #   F(c) is a singleton for each reactant c
            for e, Fc in enumerate(reactant_fates):
                if not Fc.is_singleton() :
                    log.error("Cannot condense reaction {}: ".format(reaction) + \
                            "reactant {} has multiple fates: F({}) = {}".format(
                                str(reaction.reactants[e]), str(reaction.reactants[e]), str(Fc)))
                    raise CondensationError()

            # Take all combinations of reactants and products
            new_reactant_product_combinations = it.product(
                    new_reactant_combinations, new_product_combinations)

            # Generate new reactions with each non-trivial combination
            for (reactants, products) in new_reactant_product_combinations:
                reactants = sorted(reactants)
                products = sorted(products)

                # skip the trivial case
                if reactants == products :
                    continue

                try :
                    reaction = PepperReaction(reactants, products, rtype='condensed')
                except DSDDuplicationError as e:
                    log.debug('Duplicating PepperReaction: {}'.format(e.existing))
                    reaction = e.existing

                reaction.const = self.get_condensed_rate(reaction)
                self._condensed_reactions.add(reaction)
        return

    def get_condensed_rate(self, rxn):
        """

        Condensed rate from P + Q -> Fate
        
        Sum_{i,j,l} (p_i * q_j * K_{ij}^l * d^l(Fate)

        """
        reactants = tuple(rxn.reactants)
        products = tuple(rxn.products)

        # calculate_reaction_decay_probabilities
        # reaction_decay_probabilities

        # calculate reaction rate by summing over all representative detailed reactions
        detailed_reactions_consuming = set(
                [r for reactant in reactants 
                    for c in reactant.complexes 
                    for r in self.reactions_consuming(c)])

        reaction_rate = 0.0
        for r in detailed_reactions_consuming:
            # Calculate the probability that this detailed reaction decays to the products
            # P(r -> F), where F = products.
            self.calculate_reaction_decay_probabilities(r, products)

            # probability that the products of this detailed reaction decay into fates
            # that yield the condensed reaction
            product_probability = self._reaction_decay_probabilities[(r, products)]
            assert product_probability >= 0
            assert product_probability <= 1.000001

            # probability that the resting sets comprising the reactants of
            # the condensed reaction will be in the right configuration to
            # perform the detailed reaction

            #reactant_probabilities = 1
            #for a in r.reactants:
            #    fset = frozenset(self.SCC_containing(a))
            #    rset = self._set_to_fate[fset]
            #    reactant_probabilities *= self._stationary_distributions[rset][a]

            reactant_probabilities = times(self._stationary_distributions[self._set_to_fate[
                frozenset(self.SCC_containing(a))]][a] for a in r.reactants)
            assert reactant_probabilities >= 0
            assert reactant_probabilities <= 1.000001

            # rate of the detailed reaction
            k = r.const
            assert k > 0

            # overall contribution of detailed reaction r to rate of the condensed reaction 
            reaction_rate += reactant_probabilities * k * product_probability

            if isinstance(reaction_rate, complex):
                if reaction_rate.imag > 0:
                    log.warn("Detailed reaction {} contributes a complex rate of {} + {} " + \
                         " to condensed reaction {}.".format(r, reaction_rate.real, reaction_rate.imag, rxn))

        return reaction_rate

    def get_stationary_distribution(self, scc):
        """
        Take a strongly connected component and calculate the stationary distribution.

        Returns:
            [:obj:`dict()`]: Stationary distributions: dict['cplx'] = sdist (flt)
        """
        scc_set = frozenset(scc)
        scc_list = sorted(scc)
        L = len(scc)
    
        # assign a numerical index to each complex
        complex_indices = {c: i for (i, c) in enumerate(scc_list)}
    
        # find all reactions between complexes in this SCC (non-outgoing 1,1 reactions)
        reactions = [r for c in scc for r in self.reactions_consuming(c)
                     if (r.arity == (1, 1) and not is_outgoing(r, scc_set))]
    
        # T is the transition rate matrix, defined as follows:
        # T_{ij} = { rate(j -> i)       if  i != j
        #          { - sum_{j'} T_{j'i} if  i == j
        T = np.zeros((L, L))
    
        # add transition rates for each reaction to T
        for r in reactions:
            assert len(r.reactants) == 1
            assert len(r.products) == 1
            # r : a -> b
            # T_{b,a} = rate(r : a -> b)
            a = r.reactants[0]
            b = r.products[0]
            T[complex_indices[b]][complex_indices[a]] = r.const
    
        T0 = np.copy(T)
    
        # compute diagonal elements of T
        T_diag = np.sum(T, axis=0)  # sum over columns
        for i in range(L):
            T[i][i] = -T_diag[i]
    
        # calculate eigenvalues
        (w, v) = np.linalg.eig(T)
        # w is array of eigenvalues
        # v is array of eigenvectors, where column v[:,i] is eigenvector
        # corresponding to the eigenvalue w[i].
    
        # find eigenvector corresponding to eigenvalue zero (or nearly 0)
        epsilon = 1e-5
        i = np.argmin(np.abs(w))
        if abs(w[i]) > epsilon:
            log.warn(
                ("Bad stationary distribution for resting set transition matrix. " +
                 "Eigenvalue found %f has magnitude greater than epsilon = %f. " +
                 "Markov chain may be periodic, or epsilon may be too high. Eigenvalues: %s") %
                (w(i), epsilon, str(w)))
        s = v[:, i]
    
        # check that the stationary distribution is good
        if not ((s >= 0).all() or (s <= 0).all()) : 
            log.error('Stationary distribution of resting set complex should not be an eigenvector of mixed sign. Condensed reaction rates may be incorrect.')

        s = s / np.sum(s)
        if not (abs(np.sum(s) - 1) < epsilon) :
            log.error('Stationary distribution of resting set complex should sum to 1 after normalization. Condensed reaction rates may be incorrect.')
    
        # return dict mapping complexes to stationary probabilities
        return {c: s[i] for (c, i) in complex_indices.items()}
    
    def get_exit_probabilities(self, scc):
        """
        """
        log.debug("Exit probabilities: {}".format([x.name for x in scc]))
        # build set and list of elements in SCC; assign a numerical index to each complex
        scc_set = frozenset(scc)
        scc_list = sorted(scc)
        complex_indices = {c: i for (i, c) in enumerate(scc_list)}
    
        # find all fast reactions involving complexes in this SCC
        reactions = [r for c in scc for r in self.reactions_consuming(c) if self.is_fast(r)]
    
        # sort reactions into internal and outgoing; assign a numerical index to each reaction
        r_outgoing = sorted([r for r in reactions if is_outgoing(r, scc_set)])
        r_internal = sorted([r for r in reactions if not is_outgoing(r, scc_set)])
        exit_indices = {r: i for (i, r) in enumerate(r_outgoing)}
    
        # L = # of complexes in SCC
        L = len(scc)
    
        # e = # of exit pathways
        e = len(r_outgoing)
    
        # add transition rates for each internal reaction
        T = np.zeros((L, L))
        for r in r_internal:
            assert len(r.reactants) == 1
            assert len(r.products) == 1
            a = r.reactants[0]
            b = r.products[0]
            T[complex_indices[a]][complex_indices[b]] = r.const

        # add transition rates for each outgoing reaction
        Te = np.zeros((L, e))
        for r in r_outgoing:
            a = r.reactants[0]
            Te[complex_indices[a]][exit_indices[r]] = r.const
    
        # the full transition matrix P_{L+e x L+e} would be
        #
        # P = ( Q_{LxL}  R_{Lxe}   )
        #     ( 0_{exL}  I_{exe}   )
        #
        # but we really only care about Q to calculate the fundamental matrix,
        # so we construct
        #
        # P = (T_{LxL} Te_{Lxe})
        P = np.hstack((T, Te))
    
        # then normalize P along each row, to get the overall transition
        # probabilities, e.g. P_ij = P(i -> j), where i,j in 0...L+e
        P = P / np.sum(P, 1)[:, np.newaxis]
        log.debug("P:\n{}".format(P))
    
        # extract the interior transition probabilities (Q_{LxL})
        Q = P[:, 0:L]
        log.debug("Q:\n{}".format(Q))
    
    
        # extract the exit probabilities (R_{Lxe})
        R = P[:, L:L + e]
        log.debug("R:\n{}".format(R))

    
        # calculate the fundamental matrix (N = (I_L - Q)^-1)
        N = np.linalg.inv(np.eye(L) - Q)
        log.debug("N:\n{}".format(N))

        # make sure all elements of fundamental matrix are >= 0
        if not (N >= 0).all() :  # --- commented out by EW (temporarily)
            log.error('Negative elements in fundamental matrix. Condensed reaction rates may be incorrect.')
    
        # calculate the absorption matrix (B = NR)
        B = np.dot(N, R)

        # --- added by EW as a weaker surrugate for the above, when necessary
        # assert (B >= 0).all()
    
        # return dict mapping tuples of (incoming complex, outgoing reaction)
        # to exit probabilities
        return {(c, r): B[i, j] for (c, i) in complex_indices.items()
                for (r, j) in exit_indices.items()}

class ReactionGraph(PepperCondensation):
    def __init__(self, *kargs, **kwargs):
        print('''# WARNING: peppercorn-v0.6: using depricated object name:''')
        print('''# Please rename peppercornenumerator.condense.ReactionGraph to peppercornenumerator.condense.PepperCondensation.  ''')
        super(ReactionGraph, self).__init__(*kargs, **kwargs)


class SetOfFates(object):
    """
    Corresponds to the set of multisets that can be reached from a given
    complex. Thin wrapper around the frozenset, mostly for prettier printing.
    """

    def __init__(self, states):
        # self._states = frozenset(states)
        self._states = frozenset([tuple(sorted(s)) for s in states])
        self._hash = None

    @property
    def states(self):
        return self._states

    def is_singleton(self):
        x = list(self._states)[0]
        return (len(self._states) == 1) and (len(x) == 1)

    def get_singleton(self):
        x = list(self._states)[0]
        return x[0]

    def __iter__(self):
        return iter(self.states)

    def __hash__(self):
        if(self._hash is None):
            self._hash = hash(self._states)
        return self._hash

    def __eq__(self, other):
        return sorted(list(self.states)) == sorted(list(other.states))

    def __str__(self):
        return "{ " + ", ".join(["{" + ", ".join(map(str, inner)) +
                                 "}" for inner in self.states]) + " }"

    def __len__(self):
        return len(self._states)

    def __repr__(self):
        return "SetOfFates([ " + ", ".join(
            ["(" + ", ".join(map(repr, inner)) + ")" for inner in self.states]) + " ])"


def is_outgoing(reaction, SCC_set):
    """
    Determines if the passed PepperReaction leads out of the passed SCC
    """
    if SCC_set is not set:
        SCC_set = set(SCC_set)

    # is subset of
    return not (set(reaction.products) <= SCC_set)


def times(iterable):
    """
    Multiplies each element of the given `iterable`, returns the product
    """
    return reduce(operator.mul, iterable)


def tuple_sum(iterable):
    """
    Given an iterable of tuples, concatenates each of the tuples
    and returns one big tuple. Precisely, for some list of tuples
    P = [p1, p2, p3 ...]::

        tuple_sum(P) = p1 + p2 + p3 + ...

    where + represents the concatenation of two tuples, e.g.
    (a1, a2) + (a3, a4) = (a1, a2, a3, a4).

    Example::

        tuple_sum([(1,2,3),(4,),(5,6)]) == (1, 2, 3, 4, 5, 6)

    """
    return reduce(operator.concat, tuple(iterable), tuple())


def tuple_sum_sort(iterable):
    return tuple_sort(tuple_sum(iterable))


def tuple_sort(tup):
    return tuple(sorted(tup))


def cartesian_product(iterable):
    """
    Gives the cartesian product of the passed `iterable`
    """
    # return map(tuple_sort, it.product(*iterable))
    return [tuple(x) for x in it.product(*iterable)]


def cartesian_sum(iterable):
    """
    Takes the cartesian sum over a set of multisets (counters). We define the
    cartesian sum X [+] Y between two sets X and Y as::

        X [+] Y = {x + y : x is in X and y is in Y}

    where + represents the sum of the multisets.

    Equivalently, in terms of the cartesian product X x Y, we can write X [+] Y as::

        X [+] Y = { (tuple_sum_{a in p} a) : p in X x Y }
    """
    # return map(tuple_sum,it.product(*iterable))
    return [tuple_sort(tuple_sum(x)) for x in cartesian_product(iterable)]


def get_reactions_consuming(complexes, reactions):
    """
    Returns a dict mapping each passed complex C to a list of reactions for which
    C is a reactant.
    """
    return dict((c, [r for r in reactions if (c in r.reactants)])
                for c in complexes)


def get_reactions_producing(complexes, reactions):
    """
    Returns a dict mapping each passed complex C to a list of reactions for which
    C is a product.
    """
    return dict((c, [r for r in reactions if (c in r.products)])
                for c in complexes)


def tarjans(complexes, reactions, reactions_consuming, is_fast):
    """
    Implementation of tarjans algorithm which segments SCCs of fast reactions
    """

    def strongconnect(cplx, index):
        """
        Traverse a complex and all outgoing reactions
        """
        cplx._index = index
        cplx._lowlink = index
        index += 1
        S.append(cplx)

        for reaction in reactions_consuming[cplx]:
            if reaction.arity == (1, 1) and is_fast(reaction):
                for product in reaction.products:

                    # Product hasn't been traversed; recurse
                    if product._index is None :
                        index = strongconnect(product, index)
                        cplx._lowlink = min(product._lowlink, cplx._lowlink)

                    # Product is in the current neighborhood
                    elif product in S:
                        cplx._lowlink = min(product._index, cplx._lowlink)

        if cplx._lowlink == cplx._index:
            scc = []
            while True:
                next = S.pop()
                scc.append(next)
                if next == cplx:
                    break
            SCCs.append(scc)
        return index

    index = 0
    S = []
    SCCs = []

    # Just to make sure we don't try to access an undeclared property, initialize
    # all indicies to None
    for complex in complexes:
        complex._index = None

    # Outer loop ensures that no complexes get left behind by the DFS
    for complex in complexes:
        if(complex._index is None):
            index = strongconnect(complex, index)

    return SCCs

# NOTE: temporary -- mainly a copy of PepperCondensation.get_stationary_distribution(scc)
def stationary_distribution(T, nodes = None):
    """
    Take a strongly connected component and calculate the stationary distribution.

    Args:
        T (numpy.matrix): a rate matrix 
        nodes (list, optional): A list of objects for which stationary distribution is to
            be determined.

    Returns:
        [:obj:`dict()`]: Stationary distributions: dict['cplx'] = sdist (flt)
    """
    L = len(T)
    T0 = np.copy(T)

    # compute diagonal elements of T
    T_diag = np.sum(T, axis=0)  # sum over columns
    for i in range(L):
        T[i][i] = -T_diag[i]

    # calculate eigenvalues
    (w, v) = np.linalg.eig(T)
    # w is array of eigenvalues
    # v is array of eigenvectors, where column v[:,i] is eigenvector
    # corresponding to the eigenvalue w[i].

    # find eigenvector corresponding to eigenvalue zero (or nearly 0)
    epsilon = 1e-5
    i = np.argmin(np.abs(w))
    if abs(w[i]) > epsilon:
        log.warn(
            ("Bad stationary distribution for resting set transition matrix. " +
             "Eigenvalue found %f has magnitude greater than epsilon = %f. " +
             "Markov chain may be periodic, or epsilon may be too high. Eigenvalues: %s") %
            (w(i), epsilon, str(w)))
    s = v[:, i]

    # check that the stationary distribution is good
    if not ((s >= 0).all() or (s <= 0).all()) : 
        #for x,y in zip(s, nodes):
        #    print y, '"', y.kernel_string, x
        log.error('Stationary distribution of resting set complex should not be an eigenvector of mixed sign. Condensed reaction rates may be incorrect.')

    s = s / np.sum(s)
    if not (abs(np.sum(s) - 1) < epsilon) :
        log.error('Stationary distribution of resting set complex should sum to 1 after normalization. Condensed reaction rates may be incorrect.')

    return s
    ## return dict mapping complexes to stationary probabilities
    #return {c: s[i] for (c, i) in complex_indices.iteritems()}
    



#!/usr/bin/env python
#
#  scripts/peppercorn
#  EnumeratorProject
#
from __future__ import absolute_import, division, print_function

import logging

import os
import sys
import argparse

# Import global default variables from peppercornenumerator library
import peppercornenumerator 
from peppercornenumerator import Enumerator, __version__
from peppercornenumerator.enumerator import FAST_REACTIONS
from peppercornenumerator.input import read_pil, read_seesaw, ParseException
from peppercornenumerator.reactions import branch_3way, branch_4way, opening_rate

class colors:
    RED = '\033[91m'
    YELLOW = '\033[93m'
    GREEN = '\033[92m'
    BLUE = '\033[94m'
    PINK = '\033[95m'
    CYAN = '\033[96m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    colors = [RED, YELLOW, GREEN, CYAN, BLUE, PINK]

    @staticmethod
    def color(string):
        pass

    @staticmethod
    def legend(keys=None):
        if keys is None:
            l = enumerate(colors.colors)
        else:
            l = list(zip(keys, colors.colors))
        return "\n".join([(c + str(i) + colors.ENDC) for i, c in l])

class ColorFormatter(logging.Formatter):
    def __init__(self, msg, use_color=True):
        logging.Formatter.__init__(self, msg)
        self.use_color = use_color
        self.COLORS = {
            'DEBUG': colors.CYAN,
            'INFO': colors.BLUE,
            'WARNING': colors.YELLOW,
            'ERROR': colors.RED,
            'Exception': colors.PINK,
        }
        self.RESET = colors.ENDC

    def format(self, record):
        levelname = record.levelname
        if self.use_color:
            record.levelname = self.COLORS[levelname] + \
                levelname + self.RESET
        return logging.Formatter.format(self, record)

def add_peppercorn_args(parser):
    output    = parser.add_argument_group('Peppercorn output format')
    semantics = parser.add_argument_group('Peppercorn reaction semantics')
    limits    = parser.add_argument_group('Peppercorn polymerization parameters')
    devel     = parser.add_argument_group('Peppercorn performance analysis')


    parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)
    parser.add_argument( '-v', '--verbose', action='count', default=0,
        help="Print logging output. (-vv increases verbosity.)")
    parser.add_argument('--logfile', default='', action='store', metavar='<str>',
        help="""Redirect logging information to a file.""")
    parser.add_argument('input_filename', default=None, nargs='?', metavar='<str>',
            help="Path to the input file.")

    output.add_argument('-o', '--output', dest='output_filename', default=None, metavar='<str>',
        help="""Redirect output to a file.""")
    output.add_argument('--sbml', default=None, metavar='<str>',
        help="""Print system as SBML (XML) file.""")
    output.add_argument('-c', '--condensed', action='store_true',
        help="Condense reactions into only resting complexes.")
    output.add_argument('-d', '--detailed', action='store_true',
        help="Print detailed reactions even if --condensed is chosen.")
    output.add_argument('--dry-run', action='store_true', 
        help="Dry run: read input, write output. Do not enumerate any reactions.")
    output.add_argument("--concentration-unit", default='M', action='store',
        choices=('M', 'mM', 'uM', 'nM', 'pM'),
        help="""Specify output concentration units for species and reaction rates.""")
    output.add_argument("--time-unit", default='s', action='store',
        choices=('s', 'm', 'h'),
        help="""Specify output time units for reaction rates.""")
    output.add_argument('--seesaw-explicit', action='store_true',
        help="""When using seesaw language input format, translate it into
        explicit form.""")
    output.add_argument('--seesaw-conc', default=100e-9, type=float, metavar='<flt>',
        help="""When using seesaw language input format, specify 1x
        concentrations [M].""")
    output.add_argument('--seesaw-reactions', default='', action='store',
        #choices=('seesaw', 'T20', 'T25', 'utbr', 'leak', 'reduced'),
        help=argparse.SUPPRESS)

    limits.add_argument('--max-complex-size', type=int, default=6, metavar='<int>',
        help="""Maximum number of strands allowed in a complex (used to prevent
        polymerization).""")
    limits.add_argument('--max-complex-count', type=int, default=200, metavar='<int>',
        help="""Maximum number of complexes that may be enumerated before the
        enumerator halts.""")
    limits.add_argument('--max-reaction-count', type=int, default=1000, metavar='<int>',
        help="Maximum number of reactions that may be enumerated before the enumerator halts.")

    semantics.add_argument('--k-slow', default=0.0, type=float, metavar='<flt>',
        help="Unimolecular reactions slower than this rate will be discarded.")
    semantics.add_argument('--k-fast', default=0.0, type=float, metavar='<flt>',
        help="Unimolecular reactions slower than this rate will be marked as slow.")
    semantics.add_argument('--p-min', default=0.0, type=float, metavar='<flt>',
        #help="""Minimal occupancy of a complex in steady state to engage in slow reactions.""")
        help=argparse.SUPPRESS)
    semantics.add_argument('--dG-bp', default = -1.7, type = float, metavar = '<flt>',
        help="""Adjust the average strength [kcal/mol] of a base-pair for toehold-binding 
        (affects only the opening rate). """)
    semantics.add_argument('--local-elevation', action='store_true', 
        #help="""Local probability threshold to accept an unfavorable reaction.""")
        help=argparse.SUPPRESS)

    semantics.add_argument('-L', '--release-cutoff', default=None, type=int, metavar='<int>',
        help="""Maximum number of bases that will be released spontaneously in
        an `open` reaction.""")
    semantics.add_argument('--release-cutoff-1-1', type=int, default=7, metavar='<int>',
        help="""Maximum number of bases that will be released spontaneously in
        an `open` reaction with one product.""")
    semantics.add_argument('--release-cutoff-1-2', type=int, default=7, metavar='<int>',
        help="""Maximum number of bases that will be released spontaneously in
        an `open` reaction with two products.""")

    semantics.add_argument('--no-max-helix', action='store_true', 
        help="Do not apply 'max helix' semantics.")
    semantics.add_argument('--ignore-branch-3way', action='store_true', 
        help="Ignore 3-way branch migration reactions during enumeration.")
    semantics.add_argument('--ignore-branch-4way', action='store_true',
        help="Ignore 4-way branch migration reactions during enumeration.")
    semantics.add_argument('--reject-remote', action='store_true', 
        help="""Discard remote toehold mediated 3-way and 4-way branch
        migration reactions.""")

    devel.add_argument('--interactive', action='store_true', 
        help="""Show new reactions after each step, confirm them in an
        interactive session.""")
    devel.add_argument('--interruptible', action='store_true', 
        help="""If polymerization errors are raised or after a keyboard
        interrupt, proceed with neighborhood segmentation and CRN
        condensation.""")
    devel.add_argument('--bfs-ish', action='store_true', dest='bfs',
        help="When searching for bimolecular reactions, look to the oldest complexes first.")
    devel.add_argument('--profile', action='store_true',
        help="Provide statistical profiling of enumeration (not condensation).")

    return

def set_handle_verbosity(h, v):
    if v == 0:
        h.setLevel(logging.WARNING)
    elif v == 1:
        h.setLevel(logging.INFO)
    elif v == 2:
        h.setLevel(logging.DEBUG)
    elif v >= 3:
        h.setLevel(logging.NOTSET)

def main():
    parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            description="""Peppercorn: Domain-level nucleic acid reaction enumerator.""")
    add_peppercorn_args(parser)
    args = parser.parse_args()
 
    # ~~~~~~~~~~~~~
    # Logging Setup 
    # ~~~~~~~~~~~~~
    title = "Peppercorn Domain-level Reaction Enumerator"
    logger = logging.getLogger('peppercornenumerator')
    logger.setLevel(logging.DEBUG)

    if args.logfile:
        banner = "{} {}".format(title, __version__)
        fh = logging.FileHandler(args.logfile)
        formatter = logging.Formatter('%(levelname)s - %(message)s')
        set_handle_verbosity(fh, args.verbose)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
    else:
        banner = "{} {}".format(colors.BOLD + title + colors.ENDC, 
                                  colors.GREEN + __version__ + colors.ENDC)
        ch = logging.StreamHandler()
        formatter = ColorFormatter('%(levelname)s %(message)s', use_color = True)
        set_handle_verbosity(ch, args.verbose)
        ch.setFormatter(formatter)
        logger.addHandler(ch)

    logger.info(banner)

    systeminput = args.input_filename
    if not systeminput :
        if sys.stdout.isatty():
            logger.info("Reading file from STDIN, Ctrl-D to stop")
        systeminput = ''
        for l in sys.stdin:
            systeminput += l
        if args.interactive:
            loger.error("Interactive mode needs to read input from file, not STDIN.")
            raise SystemExit

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    # Input parsing to set initial complexes for enumeration #
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    composite = None
    try :
        complexes, reactions, composite = read_pil(systeminput, 
                args.input_filename is not None, composite=True)
    except ParseException as ex_pil:
        try :
            complexes, reactions = read_seesaw(systeminput, 
                    args.input_filename is not None, 
                    conc = args.seesaw_conc, 
                    explicit = args.seesaw_explicit, 
                    reactions = args.seesaw_reactions)
        except ParseException as ex_ssw:
            logger.error('Pil-format parsing error:')
            logger.error('Cannot parse line {:5d}: "{}"'.format(ex_pil.lineno, ex_pil.line))
            logger.error('                          {} '.format(' ' * (ex_pil.col-1) + '^'))
            logger.error('SeeSaw-format parsing error:')
            logger.error('Cannot parse line {:5d}: "{}"'.format(ex_ssw.lineno, ex_ssw.line))
            logger.error('                          {} '.format(' ' * (ex_ssw.col-1) + '^'))
            raise SystemExit


    if args.dry_run:
        enum = Enumerator(list(complexes.values()), reactions)
    else:
        init_cplxs = [x for x in list(complexes.values()) if \
                x.concentration is None or float(x.concentration.value) != 0]
        enum = Enumerator(init_cplxs, reactions)

    # Log initial complexes
    logger.info("")
    logger.info("Initial complexes:")
    for c in enum.initial_complexes:
        logger.info("{}: {}".format(c, c.kernel_string))

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    # Transfer options to enumerator object #
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    logger.info("")
    logger.info("Enumeration settings:")
    enum.max_complex_size = args.max_complex_size
    logger.info("Max complex size = {}".format(enum.max_complex_size))
    enum.max_complex_count = max(args.max_complex_count, len(complexes))
    logger.info("Max complex count = {}".format(enum.max_complex_count))
    enum.max_reaction_count = max(args.max_reaction_count, len(reactions))
    logger.info("Max reaction count = {}".format(enum.max_reaction_count))
    enum.max_helix = not args.no_max_helix
    logger.info('Max-helix semantics = {}'.format(enum.max_helix))
    enum.reject_remote = args.reject_remote 
    logger.info('Reject-remote semantics = {}'.format(enum.reject_remote))
    enum.dG_bp = args.dG_bp
    logger.info('Average strength of a toehold base-pair dG_bp = {}'.format(enum.dG_bp))
    if args.ignore_branch_3way:
        if branch_3way in FAST_REACTIONS[1]:
            FAST_REACTIONS[1].remove(branch_3way)
        logger.info('No 3-way branch migration.')
    if args.ignore_branch_4way:
        if branch_4way in FAST_REACTIONS[1]:
            FAST_REACTIONS[1].remove(branch_4way)
        logger.info('No 4-way branch migration.')

    # Set either k-slow or release cutoff.
    if args.k_slow:
        if args.release_cutoff is not None:
            args.release_cutoff = None
            logger.warning('Release-cutoff overwritten by k-slow!')
        if args.release_cutoff_1_1 != args.release_cutoff_1_2:
            logger.warning('Release-cutoff (1,1) overwritten by k-slow!')
            logger.warning('Release-cutoff (1,2) overwritten by k-slow!')
        rc, k_rc = 0, None
        while True:
            rc += 1
            k_rc = opening_rate(rc)
            if k_rc < args.k_slow:
                break
        enum.release_cutoff = rc
        enum.k_slow = args.k_slow
        logger.info('Rate-dependent enumeration: k-slow = {}'.format(enum.k_slow))
        logger.info('  - corresponding release-cutoff: {} < L < {}'.format(rc-1, rc))
    else:
        if args.release_cutoff is not None:
            enum.release_cutoff = args.release_cutoff
            logger.info('Rate-independent enumeration: release cutoff L = {}'.format(
                enum.release_cutoff))
            logger.info('  - corresponding k-slow: {}'.format(opening_rate(enum.release_cutoff)))
        else:
            logger.info("Rate-independent enumeration:")
            enum.release_cutoff_1_1 = args.release_cutoff_1_1
            logger.info("  - release cutoff for reaction arity (1,1) = {}".format(
                enum.release_cutoff_1_1))
            enum.release_cutoff_1_2 = args.release_cutoff_1_2
            logger.info("  - release cutoff for reaction arity (1,2) = {}".format(
                enum.release_cutoff_1_2))
    if args.k_fast:
        enum.k_fast = args.k_fast
        logger.info('Rate-dependent enumeration: k-fast = {}'.format(enum.k_fast))

    # DEBUGGING
    enum.DFS = not args.bfs
    enum.interactive = args.interactive
    enum.interruptible = args.interruptible

    # EXPERIMENTAL
    if args.p_min != 0:
        enum.p_min = args.p_min
        logger.warning('Using experimental option: --p-min = {}'.format(enum.p_min))
    if (args.k_fast or args.k_slow) and args.local_elevation:
        logger.warning('Using experimental option: --local-elevation.')
        enum.local_elevation = True
        if enum.max_helix:
            enum.no_max_helix = False
            logger.warning('Turning off max-helix mode for local-elevation semantics.')
    elif args.local_elevation:
        logger.warning('Local-elevation and rate-independent semantics are incompatible.')

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    # Run reaction enumeration (or not) #
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    logger.info("")
    if args.profile:
        try:
            import statprof
        except ImportError as err:
            logger.warning("Python-module statprof not found, disabled Peppercorn profiling.")
            args.profile = False

    if args.dry_run:
        logger.info("Dry run (not enumerating any reactions)... ")
        enum.dry_run()
        logger.info("Done.")
    else:
        logger.info("Enumerating reactions...")
        if args.interactive:
            logger.info("Interactive mode enabled: Fast and slow reactions " + \
                        "will be printed for each complex as enumerated." + \
                        "Press ^C at any time to terminate and write accumulated" + \
                        "complexes to output.")
        if args.profile:
            statprof.start()
            try:
                enum.enumerate()
            finally:
                statprof.stop()
                statprof.display()
        else:
            enum.enumerate()
        logger.info("Done.")

    # ~~~~~~~~~~~~~~~~~~~ #
    # Handle condensation #
    # ~~~~~~~~~~~~~~~~~~~ #
    condensed = args.condensed
    detailed = (not args.condensed or args.detailed)
    if condensed:
        logger.info("Output will be condensed to remove transient complexes.")
        if args.profile:
            statprof.start()
            try:
                enum.condense()
            finally:
                statprof.stop()
                statprof.display()
        else:
            enum.condense()

    if args.sbml:
        if detailed and condensed:
            logger.error("SBML output can be detailed OR condensed, not both.")
        enum.to_sbml(args.sbml, condensed = condensed)

    output = enum.to_pil(args.output_filename, 
                         detailed=detailed, condensed=condensed, composite=composite, 
                         molarity=args.concentration_unit, time = args.time_unit)

    print(output, end='')

if __name__ == '__main__':
   main()




#!/usr/bin/env python3
#
# pilsimulator: simulate PIL files using ODEs
#
# Written by Stefan Badelt (badelt@caltech.edu).
#
# Use at your own risk.
#
#
from __future__ import absolute_import, division, print_function

import os
import sys
import argparse

from crnsimulator import ReactionGraph
from crnsimulator.odelib_template import add_integrator_args
from peppercornenumerator.input import load_pil_crn
from peppercornenumerator import __version__


def get_integrator(filename, function = 'integrate'):
    """Workaround to avoid deprecation warnings for the imp module. 

    Switch to crnsimulator.get_integrator as soon as we drop Python 2.7 support.
    """
    try: # Python 3.7
        import types
        import importlib.machinery

        loader = importlib.machinery.SourceFileLoader('myloader', filename)
        mod = types.ModuleType(loader.name)
        loader.exec_module(mod)

        return getattr(mod, function)
    except ImportError as err: # Python 2.7
        import imp
        _temp = imp.load_source('mysource', filename) # used to b odename instaed of 'mysource'
        return getattr(_temp, function)

class SimulationSetupError(Exception):
    pass

def main():
    """Translate a CRN into a system of ODEs. Optional: Simulate ODEs on-the-fly. """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    pilsim = parser.add_argument_group('pilsimulator I/O')

    pilsim.add_argument('--version', action='version', 
            version='%(prog)s ' + __version__)
    pilsim.add_argument("--dry-run", action='store_true',
            help="Do not run the simulation, only write the files.")
    pilsim.add_argument("-o", "--output", default='odesystem', metavar='<str>',
            help="Name of ODE library files.")
    pilsim.add_argument("--force", action='store_true',
            help="Overwrite existing files")
    pilsim.add_argument("--no-jacobian", action='store_true',
            help="Do not write the Jacobi-Matrix. Let the solver do the work.")

    add_integrator_args(parser)

    args = parser.parse_args()


    # ********************* #
    # ARGUMENT PROCESSING 1 #
    # ..................... #
    filename = args.output + \
        '.py' if args.output[-3:] != '.py' else args.output
    odename = 'odesystem'

    input_crn = sys.stdin.readlines()
    input_crn = "".join(input_crn)

    crn, species = load_pil_crn(input_crn)

    if crn == []:
        raise SystemExit('No Input CRN provided.')

    V = [] # sorted species (vertices) vector
    C = [] # corresponding concentration vector
    seen = set() # keep track of what species are covered

    # Move interesting species to the front, in their given order.
    labels = args.pyplot_labels
    for s in labels:
        if s in seen :
            raise SimulationSetupError('Multiple occurances of {} in labels.'.format(s))
        V.append(s)

        if species[s][0][0] != 'i': raise NotImplementedError
        C.append(species[s][1])
        seen.add(s)

    # Append the remaining specified species 
    for s in sorted(species):
        if s in seen : continue
        V.append(s)
        if species[s][0][0] != 'i':
            raise NotImplementedError('Concentrations must be given as "initial" concentrations.')
        C.append(species[s][1])
        seen.add(s)

    # Split CRN into irreversible reactions
    new = []
    for [r, p, k] in crn:
        if None in k:
            print('# Set missing rates to 1.')
            k[:] = [x if x is not None else 1 for x in k]

        if len(k) == 2:
            new.append([r, p, k[0]])
            new.append([p, r, k[1]])
        else:
            new.append([r, p, k[0]])
    crn = new

    # **************** #
    # WRITE ODE SYSTEM #
    # ................ #
    if filename != 'odesystem.py' and not args.force and os.path.exists(filename):
        print('# Reading ODE system from existing file:', filename)
    else:
        # ******************* #
        # BUILD REACTIONGRAPH #
        # ................... #
        RG = ReactionGraph(crn)
        if len(RG.species) > len(V):
            raise SimulationSetupError('Undefined species in the CRN')
        elif len(RG.species) < len(V):
            raise SimulationSetupError('Some species do not appear in the CRN:')

        # ********************* #
        # PRINT ODE TO TEMPLATE #
        # ..................... #
        filename, odename = RG.write_ODE_lib(sorted_vars = V, concvect = C,
                                             jacobian = not args.no_jacobian, 
                                             filename = filename,
                                             odename = odename)
        print('# Wrote ODE system:', filename)

    # ******************* #
    # SIMULATE ODE SYSTEM #
    # ................... #
    if args.dry_run:
        print('# Dry-run: Simulate the ODE system using:')
        print("#  python {} --help ".format(filename))
    else:
        print('# Simulating the ODE system. Change parameters using:')
        print("#  python {} --help ".format(filename))

        integrate = get_integrator(filename)

        # ********************* #
        # ARGUMENT PROCESSING 2 #
        # ..................... #
        integrate(args)

    return


if __name__ == '__main__':
    main()



#
#  peppercornenumerator/input.py
#  EnumeratorProject
#
from __future__ import absolute_import, print_function, division
from builtins import map

import logging
log = logging.getLogger(__name__)

from dsdobjects.parser import parse_pil_string, parse_pil_file
from dsdobjects.parser import parse_seesaw_string, parse_seesaw_file
from dsdobjects.parser import ParseException, PilFormatError

from peppercornenumerator.objects import PepperDomain, PepperComplex, PepperReaction, DSDDuplicationError

class InputFormatError(Exception):
    pass

def resolve_loops(loop):
    """ Return a sequence, structure pair from kernel format with parenthesis. """
    sequen = []
    struct = []
    for dom in loop :
        if isinstance(dom, str):
            sequen.append(dom)
            if dom == '+' :
                struct.append('+')
            else :
                struct.append('.')
        elif isinstance(dom, list):
            struct[-1] = '('
            old = sequen[-1]
            se, ss = resolve_loops(dom)
            sequen.extend(se)
            struct.extend(ss)
            sequen.append(old + '*' if old[-1] != '*' else old[:-1])
            struct.append(')')
    return sequen, struct

def read_reaction(line):
    rtype = line[1][0][0] if line[1] != [] and line[1][0] != [] else None
    rate = float(line[1][1][0]) if line[1] != [] and line[1][1] != [] else None
    error = float(line[1][1][1]) if line[1] != [] and line[1][1] != [] and len(line[1][1]) == 2 else None
    units = line[1][2][0] if line[1] != [] and line[1][2] != [] else None

    if rate is None:
        r = "{} -> {}".format(' + '.join(line[2]), ' + '.join(line[3]))
        log.warning("Ignoring input reaction without a rate: {}".format(r))
        return None, None, None, None, None, None
    elif rtype is None or rtype == 'condensed' or rtype not in PepperReaction.RTYPES:
        r = "{} -> {}".format(' + '.join(line[2]), ' + '.join(line[3]))
        log.warning("Ignoring input reaction of with rtype='{}': {}".format(rtype, r))
        return None, None, None, None, None, None
    else :
        r = "[{} = {:12g} {}] {} -> {}".format(
                rtype, rate, units, ' + '.join(line[2]), ' + '.join(line[3]))

    return line[2], line[3], rtype, rate, units, r

def read_pil(data, is_file = False, composite = False):
    """ Peppercorn standard input.

    Supports a variety of pil-style dialects, including kernel and enum.

    Args:
        data (str): Is either the PIL file in string format or the path to a file.
        is_file (bool): True if data is a path to a file, False otherwise
        composite (bool, optional): Returns an additional dictionary that maps
            names of composite domains (or strands) to a list of domains.
    """
    if is_file :
        parsed_file = parse_pil_file(data)
    else :
        parsed_file = parse_pil_string(data)

    domains = {'+' : '+'} # saves some code
    sequences = {}
    complexes = {}
    reactions = []
    for line in parsed_file :
        name = line[1]
        if line[0] == 'dl-domain':
            if line[2] == 'short':
                (dtype, dlen) = ('short', None)
            elif line[2] == 'long':
                (dtype, dlen) = ('long', None)
            else :
                (dtype, dlen) = (None, int(line[2]))
            if name not in domains:
                domains[name] = PepperDomain(name, dtype = dtype, length = dlen)
            cname = name[:-1] if domains[name].is_complement else name + '*'
            if cname in domains:
                assert domains[cname] == ~domains[name]
            else :
                domains[cname] = ~domains[name]

        elif line[0] == 'sl-domain':
            log.info("Ignoring sequence information for domain {}.".format(name))
            if len(line) == 4:
                if int(line[3]) != len(line[2]):
                    raise InputFormatError("Sequence/Length information inconsistent {} vs {}.".format(line[3], len(line[2])))
                domains[name] = PepperDomain(name, length = int(line[3]))
            else :
                domains[name] = PepperDomain(name, length = len(line[2]))

            domains[name].nucleotides = line[2]
            cname = name[:-1] if domains[name].is_complement else name + '*'
            if cname in domains:
                assert domains[cname] == ~domains[name]
            else :
                domains[cname] = ~domains[name]

        elif line[0] == 'composite-domain':
            # This could be a strand definition or a composite domain.
            assert name[-1] != '*'
            sequences[name] = list([domains[x] for x in line[2]])

            def comp(name):
                return name[:-1] if name[-1] == '*' else name + '*'

            sequences[comp(name)] = list([domains[comp(x)] for x in reversed(line[2])])

        elif line[0] == 'strand-complex':
            sequence = []
            for strand in line[2]:
                sequence += sequences[strand] + ['+']
            sequence = sequence[:-1]
            structure = line[3].replace(' ','')
            complexes[name] = PepperComplex(sequence, list(structure), name=name)

        elif line[0] == 'kernel-complex':
            sequence, structure = resolve_loops(line[2])

            # Replace names with domain objects.
            try :
                sequence = list([domains[d] for d in sequence])
            except KeyError:
                for e, d in enumerate(sequence):
                    if isinstance(d, PepperDomain):
                        # Happens with composite domains, see next statement e+1
                        continue
                    if d in sequences :
                        for i, c in enumerate(sequences[d]):
                            assert c.name in domains
                            if i == 0:
                                sequence[e] = c
                            else :
                                sequence.insert(e+i, c)
                                structure.insert(e+i, structure[e])

                    elif d not in domains :
                        log.warning("Assuming {} is a long domain.".format(d))
                        domains[d] = PepperDomain(d, 'long')
                        cdom = ~domains[d]
                        domains[cdom.name] = cdom
                        sequence[e] = domains[d]
                    else :
                        sequence[e] = domains[d]

            complexes[name] = PepperComplex(sequence, structure, name=name)

            if len(line) > 3 :
                assert len(line[3]) == 3
                complexes[name].concentration = tuple([line[3][0], float(line[3][1]), line[3][2]])


        elif line[0] == 'reaction':
            reactants, products, rtype, rate, units, r = read_reaction(line)
            if r is None: continue

            try :
                reactants = list([complexes[c] for c in reactants])
                products  = list([complexes[c] for c in products])
            except KeyError:
                log.warning("Ignoring input reaction with undefined complex: {}".format(r))
                continue
            
            reaction = PepperReaction(reactants, products, rtype=rtype, rate=rate)
            if reaction.rateunits != units:
                log.error("Rate units must be given in {}, not: {}.".format(reaction.rateunits, units))
                raise SystemExit
            reactions.append(reaction)

        else :
            log.warning("Ignoring {} specification: {}".format(line[0], name))
        
    if composite :
        return complexes, reactions, sequences
    else :
        return complexes, reactions

def read_kernel(data, is_file = False):
    """ New input standard, kernel notation.

    including state and reaction, ignores concetrations
    """
    log.warning("deprecated function: use read_pil")
    return read_pil(data, is_file)

def from_kernel(lines):
    """ Tranlsate a list of kernel strings. """
    log.warning('deprecated function: use read_pil') 

    # split string into lines if necessary
    if isinstance(lines, str):
        lines = lines.split("\n")

    # remove blank lines
    lines = [_f for _f in lines if _f]

    # reading pil in case of non-alphanumeric names
    complexes, _ = read_pil('\n'.join(lines))

    return (PepperDomain.MEMORY, None, complexes)

def load_pil_crn(data):
    """ Input for pilsimulator
    """
    parsed_file = parse_pil_string(data)

    sysunit = None

    species = dict()
    macrostates = dict()
    reactions = []
    for line in parsed_file:
        if line[0] == 'kernel-complex':
            name = line[1]
            conc = 0
            if len(line) > 3 :
                init = line[3][0]
                conc = float(line[3][1])
                if sysunit is None:
                    sysunit = line[3][2]
                else :
                    if sysunit != line[3][2]:
                        raise PilFormatError(
                                'Conflicting units: {} vs. {}'.format(sysunit, line[3][2]))
                if init[0] != 'i':
                    raise NotImplementedError('concentrations must be specified as *initial*')
            species[name] = ('initial', conc)
        elif line[0] == 'resting-macrostate':
            name = line[1]
            conc = ['initial', 0]
            for sp in line[2]:
                assert sp in species
                if species[sp][0][0] != 'i':
                    raise NotImplementedError('concentrations must be specified as *initial*')
                conc[1] += species[sp][1]
            macrostates[name] = tuple(conc)
        elif line[0] == 'reaction':
            info = line[1]
            reactants = line[2]
            products = line[3]
            assert len(info) == 3
            rate = float(info[1][0])
            if sysunit and [x for x in info[2][0].split('/')[1:-1] if x != sysunit]:
                raise PilFormatError('Conflicting units: {} vs. {}'.format(sysunit, info[2][0]))
            reactions.append([reactants, products, [rate]])
        elif line[0] == 'dl-domain' :
            pass
        else :
            print('# Ignoring Keyword: {}'.format(line[0]))

    crnspecies = dict()
    condensed = bool(macrostates)
    for [reac,prod,rate] in reactions:
        for sp in reac + prod:
            if condensed:
                # If macrostates are specified, then they must contain all
                # species of the CRN.. otherwise, we'd have to make sure that
                # we don't return a mix of detailed and condensed CRN.
                if sp not in macrostates:
                    raise InputFormatError('Undefined macrostate in condensed input CRN: {}'.format(sp))
                crnspecies[sp] = macrostates[sp]
            elif sp not in crnspecies:
                crnspecies[sp] = species.pop(sp, ('initial', 0))

    if not condensed and bool(species):
        log.warning("Some species do not appear in the CRN: {}".format(list(species.keys())))

    return reactions, crnspecies

def read_seesaw(data, is_file = False, conc = 100e-9, explicit = True, reactions = ''):
    """Translates a seesaw circuit into PepperComplexes.

    Args:
        is_file (bool, optional): Pareses a file if True, parses a string if False
        conc (float, optional): The 1x concentration value in M.    
            Defaults to 100 nM
        explicit (bool, optional): Translate to the exact explicit notation, or to 
            the easier shorthand-notation. Defaults to True: explicit.

    The explicit notation is more accurate, but also harder to enumerate.
    Typical domain lengths:
        gate-b: 5 + 15 + 5
        thld-b: 5 + 5 + 15
        long-domain: 10 + 5
        toehold: 5
        gate-top: 15 + 5 + 15 != 32!!
 
    """

    if is_file :
        parsed_file = parse_seesaw_file(data)
    else :
        parsed_file = parse_seesaw_string(data)

    domains = {'+' : '+'}
    def assgn_domain(name, l):
        if name not in domains:
            domains[name] = PepperDomain(name, length = l)
            cname = name[:-1] if domains[name].is_complement else name + '*'
            domains[cname] = ~domains[name]
        return domains[name]

    if explicit:
        toe = assgn_domain('toe', 3) # toehold TCT - AGA
        clp = assgn_domain('clp', 2) # clamp   CA  - TG
        dlen = 6
        tlen = 5
    else :
        toe = assgn_domain('toe', 5) # toehold CATCT - AGATG
        dlen = 10 # length of the domain
        tlen =  5 # lenth of the threshold

    complexes = {}
    def make_wire(w, name=None):
        """Translates a wire into a PepperComplex.

        Every complex gets 0 M concentration as default, i.e. it is excluded
        from enumeration.

        Args:
            name (str, optional): Provida a name for the complex that is
                different from the internal wire representation.

        explicit:
            ['w', ['1', '2']] -> w1_2 = c s2 c t c s1 c
        shorthand:
            ['w', ['1', '2']] -> w1_2 = s2 t s1

        """
        if w[0] != 'w': 
            raise InputFormatError('Does not look like a wire: {}'.format(w))

        [nI, nO] = w[1]
        wire = 'w{}_{}'.format(nI, nO)
        if wire not in complexes:
            dIa = assgn_domain('s{}a'.format(nI), l=dlen)
            dIt = assgn_domain('s{}t'.format(nI), l=tlen)
            dOa = assgn_domain('s{}a'.format(nO), l=dlen)
            dOt = assgn_domain('s{}t'.format(nO), l=tlen)
            if explicit:
                sequence = [clp, dOt, dOa, clp, toe, clp, dIt, dIa, clp]
                structure = list('.........')
            else:
                sequence = [dOt, dOa, toe, dIt, dIa]
                structure = list('.....')
            complexes[wire] = PepperComplex(sequence, structure, 
                    name=name if name else wire)
            complexes[wire].concentration = ('i', 0, 'M')
        return complexes[wire]

    def make_thld(t):
        """ ['t', [inp, name]] """
        if t[0] != 't': 
            raise InputFormatError('Does not look like a threshold: {}'.format(t))
        [nI, nO] = t[1]
        thld = 'T{}_{}'.format(nI, nO)
        waste1 = 'Waste{}_{}'.format(nI, nO)
        waste2 = "Waste_{}".format(nO)
        if thld not in complexes:
            dIa = assgn_domain('s{}a'.format(nI), l=dlen)
            dIt = assgn_domain('s{}t'.format(nI), l=tlen)
            dOa = assgn_domain('s{}a'.format(nO), l=dlen)
            dOt = assgn_domain('s{}t'.format(nO), l=tlen)
            if explicit:
                sequence = [clp, dOt, dOa, clp, '+', ~dIt, ~clp, ~toe, ~clp, ~dOa, ~dOt, ~clp]
                structure = list('((((+...))))')
            else :
                sequence = [dOt, dOa, '+', ~dIt, ~toe, ~dOa, ~dOt]
                structure = list('((+..))')
            complexes[thld] = PepperComplex(sequence, structure, name=thld)
            complexes[thld].concentration = ('i', 0, 'M')

            # Name the waste as well, for convenience...
            assert waste1 not in complexes
            if explicit:
                sequence = [clp, dOt, dOa, clp, toe, clp, dIt, dIa, clp, '+', ~dIt, ~clp, ~toe, ~clp, ~dOa, ~dOt, ~clp]
                structure = list('(((((((..+)))))))')
            else :
                sequence = [dOt, dOa, toe, dIt, dIa, '+', ~dIt, ~toe, ~dOa, ~dOt]
                structure = list('((((.+))))')
            complexes[waste1] = PepperComplex(sequence, structure, name=waste1)
            complexes[waste1].concentration = ('i', 0, 'M')

            if waste2 not in complexes:
                if explicit:
                    raise NotImplementedError
                else :
                    complexes[waste2] = PepperComplex([dOt, dOa], ['.','.'], name=waste2)
                    complexes[waste2].concentration = ('i', 0, 'M')

        return complexes[thld], complexes[waste1], complexes[waste2]

    def make_gate(g):
        """ Two options: 
        1) ['g', [['w', ['31', '25']], '25']]
        2) ['g', ['31', ['w', ['31', '25']]]]
        """
        if g[0] != 'g': 
            raise InputFormatError('Does not look like a gate: {}'.format(g))

        if isinstance(g[1][0], list):
            # ['g', [['w', ['1', '2']], '2']]
            [[_, [nI, nO]], nG] = g[1]
            assert _ == 'w' and nG == nO
        elif isinstance(g[1][1], list):
            # ['g', ['1', ['w', ['1', '2']]]]
            [nG, [_, [nI, nO]]] = g[1]
            assert _ == 'w' and nG == nI

        if nI == nG : # this is a producing gate:
            assert nO != nG
            dGa = assgn_domain('s{}a'.format(nG), l=dlen)
            dGt = assgn_domain('s{}t'.format(nG), l=tlen)
            dOa = assgn_domain('s{}a'.format(nO), l=dlen)
            dOt = assgn_domain('s{}t'.format(nO), l=tlen)

            wire = 'w{}_{}'.format(nG, nO) # output wire
            gate = 'G_g{}_{}'.format(nG, wire)
            if explicit:
                sequence  = [clp, dOt, dOa, clp, toe, clp, dGt, dGa, clp, '+', 
                            ~clp, ~toe, ~clp, ~dGa, ~dGt, ~clp, ~toe, ~clp]
                structure = list('...((((((+..))))))')
            else :
                sequence  = [dOt, dOa, toe, dGt, dGa, '+', ~toe, ~dGa, ~dGt, ~toe]
                structure = list('..(((+.)))')

        else : # this is a consumed gate:
            assert nO == nG and nI != nG
            dIa = assgn_domain('s{}a'.format(nI), l=dlen)
            dIt = assgn_domain('s{}t'.format(nI), l=tlen)
            dGa = assgn_domain('s{}a'.format(nG), l=dlen)
            dGt = assgn_domain('s{}t'.format(nG), l=tlen)

            wire = 'w{}_{}'.format(nI, nG) # output wire
            gate = 'G_{}_g{}'.format(wire, nG)
            if explicit:
                sequence  = [clp, dGt, dGa, clp, toe, clp, dIt, dIa, clp, '+', 
                            ~clp, ~toe, ~clp, ~dGa, ~dGt, ~clp, ~toe, ~clp]
                structure = list('((((((...+))))))..')
            else :
                sequence  = [dGt, dGa, toe, dIt, dIa, '+', ~toe, ~dGa, ~dGt, ~toe]
                structure = list('(((..+))).')
 
        if gate not in complexes:
            complexes[gate] = PepperComplex(sequence, structure, name=gate)
            complexes[gate].concentration = ('i', 0, 'M')
        return complexes[gate]

    for line in parsed_file :
        if line[0] == 'INPUT':
            # use the given name if it is not a digit...
            name = None if line[1][0].isdigit() else line[1][0]
            w = make_wire(line[2], name)
            assert w.concentration.value == 0
            w.concentration = None

        elif line[0] == 'OUTPUT':
            # use the given name if it is not a digit...
            name = None if line[1][0].isdigit() else line[1][0]
            if line[2][0] == 'w':
                make_wire(line[2], name)
            elif line[2][0] == 'Fluor':
                nF = line[2][1]
                fluor = "F_{}".format(line[2][1])
                name = fluor if name is None else name
                dFa = assgn_domain('s{}a'.format(nF), l=dlen)
                dFt = assgn_domain('s{}t'.format(nF), l=tlen)
                if fluor not in complexes:
                    if explicit:
                        complexes[fluor] = PepperComplex([clp, dFt, dFa, clp], ['.','.','.','.'], name=name)
                    else :
                        complexes[fluor] = PepperComplex([dFt, dFa], ['.','.'], name=name)
                    complexes[fluor].concentration = ('i', 0, 'M')
            else:
                raise InputFormatError('Unknown output format: {}'.format(line))

        elif line[0] == 'seesaw':
            # seesaw[5,{2},{6,7}]
            # rxn w(2-5) + g(5,5-6) <=> w(5-6) + g(2-5,5)
            #
            # produces all species for a custom seesaw gate, but with ZERO
            # concentration. If no concentrations are specified later via conc
            # statements), then this species is initially not present.
            #
            [gate, inp, out] = line[1]
            for d in inp:
                make_wire(['w', [d, gate]])                 # input wire
                make_thld(['t', [d, gate]])                 # threshold cplx (consuming input)
                make_gate(['g', [['w', [d, gate]], gate]])  # consumed gate (producing input)

            for d in out:
                make_wire(['w', [gate, d]])                 # output wire
                make_gate(['g', [gate, ['w',[gate, d]]]])   # active gate (producing output)

        elif line[0] in ('seesawOR', 'seesawAND'):
            # seesaw[2,5,{1,4},{3,9,10}]
            [g1, g2, inp, out] = line[1]
            n = len(inp)
            m = len(out)

            # make gates for all different inputs
            for d in inp:
                Iw = make_wire(['w', [d, g1]])
                Ig = make_gate(['g', [g1, ['w', [g1, g2]]]])

                Ow = make_wire(['w', [g1, g2]])
                Og = make_gate(['g', [['w', [d, g1]], g1]]) # consumed gate

            # Thresholding reaction
            Iw = make_wire(['w',[g1, g2]])
            Th,w1,w2 = make_thld(['t',[g1, g2]]) 
            if line[0] == 'seesawOR':
                assert Th.concentration.value == 0
                Th.concentration = ('i', 1.1 * 0.6 * conc, 'M')
            else:
                assert Th.concentration.value == 0
                Th.concentration = ('i', 1.1 * (n-1 + 0.2) * conc, 'M')

            # Connection reaction (assign intial concentration)
            Ig = make_gate(['g', [g1, ['w', [g1, g2]]]])
            assert Ig.concentration.value == 0
            Ig.concentration = ('i', n * conc, 'M')

            # fanout for all outputs
            for d in out + ['f']:
                Iw = make_wire(['w',[g1, g2]])
                Ig = make_gate(['g', [g2, ['w',[g2, d]]]])

                Ow = make_wire(['w', [g2, d]])
                Og = make_gate(['g', [['w', [g1, g2]], g2]]) #consumed gate

                if d == 'f':
                    assert Ow.concentration.value == 0
                    fmod = Ow.concentration.mode
                    fcon = Ow.concentration.value
                    funi = Ow.concentration.unit
                    fcon += 2 * m * conc
                    Ow.concentration = (fmod, fcon, funi)
                else :
                    Ig.concentration = ('i', conc, 'M')

        elif line[0] == 'inputfanout':
            [g1, inp, out] = line[1]
            n = len(inp)
            m = len(out)

            make_wire(['w', [inp, g1]])
            cx,_,_ = make_thld(['t', [inp, g1]])
            assert cx.concentration.value == 0
            cx.concentration = ('i', 1.1 * 0.2 * conc, 'M')
            make_gate(['g', [['w',[inp, g1]], g1]])

            for d in out + ['f']:
                cxw = make_wire(['w', [g1, d]])
                cxg = make_gate(['g', [g1, ['w', [g1, d]]]])
                if d == 'f':
                    assert cxw.concentration.value == 0
                    cxwmod = Ow.concentration.mode
                    cxwcon = Ow.concentration.value
                    cxwuni = Ow.concentration.unit
                    cxwcon += 2 * m * conc
                    cxw.concentration = (cxwmod, cxwcon, cxwuni)
                else :
                    assert cxg.concentration.value == 0
                    cxg.concentration = ('i', conc, 'M')

        elif line[0] == 'reporter':
            #['reporter', ['25', '31']]
            [nR, nI] = line[1]
            sw = make_wire(['w', [nI, nR]])

            rep = 'R_{}'.format(nR)
            assert rep not in complexes

            dRa = assgn_domain('s{}a'.format(nR), l=dlen)
            dRt = assgn_domain('s{}t'.format(nR), l=tlen)
            if explicit:
                sequence  = [clp, dRt, dRa, clp, '+', ~clp, ~toe, ~clp, ~dRa, ~dRt, ~clp]
                structure = list('((((+..))))')
            else: 
                sequence  = [dRt, dRa, '+', ~toe, ~dRa, ~dRt]
                structure = list('((+.))')
            complexes[rep] = PepperComplex(sequence, structure, name=rep)
            complexes[rep].concentration = ('i', 1.5 * conc, 'M')

            # TODO: swap F and Q!!!
            flr = "F_{}".format(nR)
            if flr not in complexes:
                if explicit:
                    complexes[flr] = PepperComplex([clp, dRt, dRa, clp], 
                                                   ['.','.','.','.'], name=flr)
                else :
                    complexes[flr] = PepperComplex([dRt, dRa], ['.','.'], name=flr)
                complexes[flr].concentration = ('i', 0, 'M')

            que = "Q_{}".format(nR)
            if que not in complexes:
                dIa = assgn_domain('s{}a'.format(nI), l=dlen)
                dIt = assgn_domain('s{}t'.format(nI), l=tlen)
                if explicit:
                    raise NotImplementedError
                else :
                    sequence  = [dRt, dRa, toe, dIt, dIa, '+', ~toe, ~dRa, ~dRt]
                    structure = list('(((..+)))')
                    complexes[que] = PepperComplex(sequence, structure, name=que)
                    complexes[que].concentration = ('i', 0, 'M')

        elif line[0] == 'conc':
            if line[1][0] == 'w':
                cx = make_wire(line[1])
            elif line[1][0] == 'g':
                cx = make_gate(line[1])
            elif line[1][0] == 'th':
                tmp  = line[1][1][0]
                tmp[0] = 't'
                cx,_,_ = make_thld(tmp)
            else:
                print(line)
                raise NotImplementedError
            # input wires have None at default
            assert cx.concentration is None or cx.concentration.value == 0
            cx.concentration = ('i', conc*float(line[2]), 'M')

        else :
            log.warning('Keyword not supported: {}'.format(line[0]))

    if reactions: # e.g. 'seesaw-T25-utbr-leak-reduced'
        for term in reactions.split('-'):
            assert term in ('seesaw', 'T20', 'T25', 'utbr', 'leak', 'reduced')
        T = 25 if 'T25' in reactions else 20
        utbr = True if 'utbr' in reactions else False
        leak = True if 'leak' in reactions else False
        reduced = True if 'reduced' in reactions else False
        reactions = get_seesaw_compiler_reactions(complexes, T, utbr, leak, reduced)
    else :
        reactions = []

    return complexes, reactions

def get_seesaw_compiler_reactions(complexes, T=20, utbr=False, leak=False, global_species=False):
    """Write the standard seesaw reactions. 

    Takes a dictionary of seesaw complexes and finds all designed and side
    reactions described in Qian & Winfree 2011, Supplemental Section S13

    Args:
        complexes (dict, optional): Maps seesaw names to complexes. Note, if
            complexes got renamed due to INPUT/OUTPUT statements, then
            complex.name is not the same as the dictionary key.
        T (int, optional): Temperature, can be either 20 or 25 *C.
        utbr (bool, optional): Return universal toehold binding reactions.
        leak (bool, optional): Return leak reactions.
        global_species (bool, optional): Use global Threshold, Wire and Gate
            species to reduce the number of explicit toehold binding reactions.
    
    """
    PepperReaction.RTYPES.add('seesaw')      # desinged seesawing
    PepperReaction.RTYPES.add('seesaw-thld') # designed thresholding
    PepperReaction.RTYPES.add('seesaw-rep')  # designed reporting
    PepperReaction.RTYPES.add('side-utbr')   # side reaction: universal toehold binding 
    PepperReaction.RTYPES.add('side-leak')   # side reaction: leak

    ks  = 5e4 # /M/s
    kf  = 2e6 # /M/s
    if T == 20:
        krs = 0.5 # /s
        krf = 10  # /s
        kl  = 1   # /M/s
    elif T == 25:
        krs = 1.3 # /s
        krf = 26  # /s
        kl  = 10  # /M/s
    else :
        raise NotImplementedError('Please use temperature 20 or 25.')

    # complexes[k] = PepperComplex ... k != complex.name 
    wires = []
    gates = []
    thlds = []
    freps = []
    for k,x in list(complexes.items()):
        if k[0] == 'w':
            x.seesawname = k
            wires.append(x)
        elif k[0]=='G': 
            gates.append(x)
        elif k[0]=='R':
            freps.append(x)
        elif k[0]=='T':
            thlds.append(x)

    count=0
    reactions = []
    
    if utbr and global_species :
        # Make the universal wire W
        # This is the fuel value... 
        conc = sum([x.concentration.value if x.concentration else 0 for x in wires])
        if conc:
            uW = PepperDomain('dummy_{}'.format('W'), dtype='long')
            uW = PepperComplex(sequence=[uW], structure=['.'], name='W')
            uW.concentration = ('initial', conc, 'M')
            complexes[uW.name] = uW

        # Make the universal gate G (Gates and Reporters)
        uG = PepperDomain('dummy_{}'.format('G'), dtype='long')
        uG = PepperComplex(sequence=[uG], structure=['.'], name='G')
        conc = sum([x.concentration.value if x.concentration else 0 for x in gates + freps])
        uG.concentration = ('initial', conc, 'M')
        complexes[uG.name] = uG

        # Make the universal threshold TH 
        uT = PepperDomain('dummy_{}'.format('TH'), dtype='long')
        uT = PepperComplex(sequence=[uT], structure=['.'], name='TH')
        conc = sum([x.concentration.value if x.concentration else 0 for x in thlds])
        uT.concentration = ('initial', conc, 'M')
        complexes[uT.name] = uT

    for w in wires:
        wi, wo = w.seesawname[1:].split('_')

        # Find all reactions with gates:
        tot_G_conc = 0
        for g in gates:

            G, gn, gi, go = str(g).split('_')
            if gn[0] == 'g': # a forward gate G5:5,6
                assert gi[0] == 'w'
                gw = complexes[gi+'_'+go]
                G = '+'
            else : # a reverse gate G5,6:6
                assert gn[0] == 'w'
                assert go[0] == 'g'
                [gi, go, gn] = [gn, gi, go]
                gw = complexes[gi+'_'+go]
                G = '-'

            gwi, gwo = gw.seesawname[1:].split('_')
            gn = gn[1:]

            if G == '+' and wo == gn:
                # a regular interaction
                og = complexes['G_' + w.seesawname + '_' + 'g' + gn]
                reactions.append(PepperReaction([w, g], [gw, og], rtype='seesaw', rate=ks))
                #print("reaction [seesaw = {} /M/s ] {} + {} -> {} + {}".format(ks, w, g, gw, og))
                continue

            if G == '-' and wi == gn:
                # a regular interaction
                og = complexes['G_' + 'g' + gn + '_' + w.seesawname]
                reactions.append(PepperReaction([w, g], [gw, og], rtype='seesaw', rate=ks))
                #print("reaction [seesaw = {} /M/s ] {} + {} -> {} + {}".format(ks, w, g, gw, og))
                continue

            if leak and G == '+' and wi == gwi:# and wo != gwo:
                og = complexes['G_' + 'g' + gn + '_' + w.seesawname]
                reactions.append(PepperReaction([w, g], [gw, og], rtype='side-leak', rate=kl))
                #print("reaction [seesaw-leak = {} /M/s ] {} + {} -> {} + {}".format(kl,w,g,gw, og))

            if leak and G == '-' and wo == gwo:# and wi != gwi:
                og = complexes['G_' + w.seesawname + '_' + 'g' + gn]
                reactions.append(PepperReaction([w, g], [gw, og], rtype='side-leak', rate=kl))
                #print("reaction [seesaw-leak = {} /M/s ] {} + {} -> {} + {}".format(kl,w,g,gw, og))

            if utbr and global_species:
                ## TODO: check if that makes sense...?
                if w.concentration and w.concentration.value != 0:
                    try : # all Wires together
                        gWn = g.name + '_W'
                        gW = PepperDomain('dummy_{}'.format(gWn), dtype='long')
                        gW = PepperComplex(sequence=[gW], structure=['.'], name=gWn)
                        gW.concentration = ('initial', 0, 'M')
                        complexes[gWn] = gW

                        reactions.append(PepperReaction([g, uW], [gW], rtype='side-utbr', rate=kf))
                        reactions.append(PepperReaction([gW], [g, uW], rtype='side-utbr', rate=krf))
                    except DSDDuplicationError as err:
                        pass
                else :
                    try: # all Gates together
                        wGn = w.name + '_G'
                        wG = PepperDomain('dummy_{}'.format(wGn), dtype='long')
                        wG = PepperComplex(sequence=[wG], structure=['.'], name=wGn)
                        wG.concentration = ('initial', 0, 'M')
                        complexes[wGn] = wG

                        reactions.append(PepperReaction([w, uG], [wG], rtype='side-utbr', rate=kf))
                        reactions.append(PepperReaction([wG], [w, uG], rtype='side-utbr', rate=krf))
                    except DSDDuplicationError as err:
                        pass

            elif utbr :
                count += 1
                a = PepperDomain('dummy_{}'.format(count), dtype='long')
                a = PepperComplex(sequence=[a], structure=['.'], prefix='s')
                complexes[a.name] = a
                reactions.append(PepperReaction([w, g], [a], rtype='side-utbr', rate=kf))
                reactions.append(PepperReaction([a], [w, g], rtype='side-utbr', rate=krf))
        
        for g in freps:
            # R_6
            if '_{}'.format(wo) in str(g):
                # a regular interaction
                of = complexes['F_'+wo]
                oq = complexes['Q_'+wo]
                reactions.append(PepperReaction([w, g], [of, oq], rtype='seesaw-rep', rate=ks))
                #print("reaction [seesaw = {} /M/s ] {} + {} -> {} + {}".format(ks, w, g, of, oq))
                continue

            if utbr and global_species:
                # TODO: check if that makes sense...?
                if w.concentration and w.concentration.value != 0:
                    try: # all-wires product
                        gWn = g.name + '_W'
                        gW = PepperDomain('dummy_{}'.format(gWn), dtype='long')
                        gW = PepperComplex(sequence=[gW], structure=['.'], name=gWn)
                        gW.concentration = ('initial', 0, 'M')
                        complexes[gWn] = gW

                        reactions.append(PepperReaction([g, uW], [gW], rtype='side-utbr', rate=kf))
                        reactions.append(PepperReaction([gW], [g, uW], rtype='side-utbr', rate=krf))
                    except DSDDuplicationError as err:
                        pass
                else:
                    try: # all Gates together
                        wGn = w.name + '_G'
                        wG = PepperDomain('dummy_{}'.format(wGn), dtype='long')
                        wG = PepperComplex(sequence=[wG], structure=['.'], name=wGn)
                        wG.concentration = ('initial', 0, 'M')
                        complexes[wGn] = wG

                        reactions.append(PepperReaction([w, uG], [wG], rtype='side-utbr', rate=kf))
                        reactions.append(PepperReaction([wG], [w, uG], rtype='side-utbr', rate=krf))
                    except DSDDuplicationError as err:
                        pass

            elif utbr:
                count+=1
                a = PepperDomain('dummy_{}'.format(count), dtype='long')
                a = PepperComplex(sequence=[a], structure=['.'], prefix='s')
                complexes[a.name] = a
                reactions.append(PepperReaction([w, g], [a], rtype='side-utbr', rate=kf))
                reactions.append(PepperReaction([a], [w, g], rtype='side-utbr', rate=krf))
                #print("reaction [seesaw-rep = {} /M/s ] {} + {} -> {}".format(kf,w,g,a))
                #print("reaction [seesaw-rep = {} /s ] {} -> {} + {}".format(krf,a,w,g))

        for g in thlds:
            # T2_5
            if 'T{}_{}'.format(wi, wo) == str(g):
                tc = complexes['Waste{}_{}'.format(wi, wo)]
                tw = complexes['Waste_'+wo]
                reactions.append(PepperReaction([w, g], [tc, tw], rtype='seesaw-thld', rate=kf))
                #print("reaction [seesaw = {} /M/s ] {} + {} -> {}_c + W_{}".format(kf, w, g, g, wo))
                continue

            if utbr and global_species:
                # TODO: check if that makes sense...?
                if w.concentration and w.concentration.value != 0:
                    try: # all Wires together
                        gWn = g.name + '_W'
                        gW = PepperDomain('dummy_{}'.format(gWn), dtype='long')
                        gW = PepperComplex(sequence=[gW], structure=['.'], name=gWn)
                        gW.concentration = ('initial', 0, 'M')
                        complexes[gWn] = gW

                        reactions.append(PepperReaction([g, uW], [gW], rtype='side-utbr', rate=kf))
                        reactions.append(PepperReaction([gW], [g, uW], rtype='side-utbr', rate=krs))
                    except DSDDuplicationError as err:
                        pass
                else:
                    try: # all Gates together
                        wGn = w.name + '_TH'
                        wG = PepperDomain('dummy_{}'.format(wGn), dtype='long')
                        wG = PepperComplex(sequence=[wG], structure=['.'], name=wGn)
                        wG.concentration = ('initial', 0, 'M')
                        complexes[wGn] = wG

                        reactions.append(PepperReaction([w, uT], [wG], rtype='side-utbr', rate=kf))
                        reactions.append(PepperReaction([wG], [w, uT], rtype='side-utbr', rate=krs))
                    except DSDDuplicationError as err:
                        pass

            elif utbr:
                count+=1
                a = PepperDomain('dummy_{}'.format(count), dtype='long')
                a = PepperComplex(sequence=[a], structure=['.'], prefix='s')
                complexes[a.name] = a
                reactions.append(PepperReaction([w, g], [a], rtype='side-utbr', rate=kf))
                reactions.append(PepperReaction([a], [w, g], rtype='side-utbr', rate=krs))
                #print("reaction [seesaw-thd = {} /M/s ] {} + {} -> {}".format(kf,w,g,a))
                #print("reaction [seesaw-thd = {} /s ] {} -> {} + {}".format(krs,a,w,g))

    return reactions



#
#  peppercornenumerator/output.py
#  EnumeratorProject
#
from __future__ import absolute_import, print_function, division
from builtins import map

import logging
log = logging.getLogger(__name__)

import xml.dom.minidom 
from collections import Counter

from peppercornenumerator import __version__
from peppercornenumerator.utils import natural_sort
from peppercornenumerator.utils import PeppercornUsageError
from peppercornenumerator.condense import PepperCondensation

def write_pil(enumerator, fh = None, detailed = True, condensed = False, 
        composite = None, molarity = 'M', time = 's'):
    """Write the contents of :obj:`Enumerator()` into a *.pil file.

    Args:
      enumerator (:obj:`Enumerator()`): The enumertor object.
      fh (filehandle): The filehandle to which the output is written to. 
        If fh is None, then output is written to a string and returned.
      detailed (bool, optional): Print detailed CRN. Defaults to True.
      condensed (bool, optional): Print condensed CRN. Defaults to False.
      composite (dict, optional): A dictionary of composite domains (or
        strands) to be included into the *.pil file.
    """

    out = []
    def output_string(string):
        if fh is None:
            out.append(string)
        else :
            fh.write(string)

    output_string("# File generated by peppercorn-{}\n".format(__version__))

    # Print domains
    seen = set()
    output_string("\n# Domains ({}) \n".format(len(enumerator.domains)))
    for dom in natural_sort(enumerator.domains):
        if dom.is_complement and not dom.nucleotides: 
            dom = ~dom
        if dom not in seen :
            if dom.nucleotides:
                output_string("sequence {:s} = {} : {:d}\n".format(
                    dom.name, dom.nucleotides, len(dom)))
            else :
                output_string("length {:s} = {:d}\n".format(dom.name, len(dom)))
            seen.add(dom)

    if composite:
        output_string("\n# Strands or composite domains ({}) \n".format(len(composite)))
        for comp in composite:
            if comp[-1]=='*' and comp[:-1] in composite:
                continue
            output_string("sup-sequence {} = {} : {:d}\n".format(comp, ' '.join(
                map(str, composite[comp])), sum(map(len,composite[comp]))))

    # Print resting complexes
    output_string("\n# Resting complexes ({}) \n".format(len(
        enumerator.resting_complexes)))
    for cplx in natural_sort(enumerator.resting_complexes):
        if cplx.concentration :
            output_string("{:s} = {:s} @{} {} {}\n".format(cplx.name, 
                cplx.kernel_string, *cplx.concentrationformat(molarity)))
        else:
            output_string("{:s} = {:s}\n".format(cplx.name, cplx.kernel_string))
 
    if condensed :
        # Print resting macrostates
        output_string("\n# Resting macrostates ({}) \n".format(
            len(enumerator.resting_macrostates)))
        for resting in natural_sort(enumerator.resting_macrostates):
            output_string("macrostate {:s} = [{:s}]\n".format(str(resting), 
                ', '.join(map(str,resting.complexes))))

        # Print reactions
        output_string("\n# Condensed reactions ({}) \n".format(len(enumerator.condensed_reactions)))
        for rxn in natural_sort(enumerator.condensed_reactions):
            output_string("reaction {:s}\n".format(rxn.full_string(molarity, time)))

    if detailed :
        # Print transient complexes
        output_string("\n# Transient complexes ({}) \n".format(len(enumerator.transient_complexes)))
        for cplx in natural_sort(enumerator.transient_complexes):
            if cplx._concentration :
                output_string("{:s} = {:s} @{} {} {}\n".format(cplx.name, cplx.kernel_string, 
                    *cplx.concentrationformat(molarity)))
            else:
                output_string("{:s} = {:s}\n".format(cplx.name, cplx.kernel_string))

        # Print reactions
        output_string("\n# Detailed reactions ({}) \n".format(len(enumerator.reactions)))
        for rxn in natural_sort(enumerator.reactions):
            output_string("reaction {:s}\n".format(rxn.full_string(molarity, time)))

    return ''.join(out)

def write_crn(enumerator, crn, condensed = False, molarity = 'M', time = 's'):
    crn.write("# File generated by peppercorn-{}\n".format(__version__))

    if condensed :
        # Print reactions
        crn.write("\n# Condensed reactions: concentration = {}, time = {}\n".format(
            molarity, time))
        for reaction in natural_sort(enumerator.condensed_reactions):
            units = [molarity] * (reaction.arity[0] - 1) + [time]
            rate = reaction.rateformat(units)
            crn.write("{} [k = {}]\n".format(reaction, rate.constant))
    else :
        # Print reactions
        crn.write("\n# Detailed reactions: concentration = {}, time = {}\n".format(
            molarity, time))
        for reaction in natural_sort(enumerator.reactions):
            units = [molarity] * (reaction.arity[0] - 1) + [time]
            rate = reaction.rateformat(units)
            crn.write("{} [k = {}]\n".format(reaction, rate.constant))

def write_vdsd(enumerator, fh = None, detailed = True, condensed = False,
        composite = None, molarity = 'nM', time = 's'):
    """ Write an Enumerator Object into VisualDSD *.crn format.

    directive simulation {plots=[sp_0; sp_1; sp_2; sp_3; sp_4; sp_5; sp_6; sp_7; sp_8; sp_9; sp_10]; }

    | 10 sp_0
    | 100 sp_1
    | sp_1 <->{displace}{displace} sp_10
    | sp_9 ->{displace} sp_7 + sp_6
    | sp_0 + sp_1 <->{bind}{unbind} sp_9
    | sp_9 <->{displace}{displace} sp_8
    | sp_0 + sp_10 <->{bind}{unbind} sp_8
    | sp_5 ->{displace} sp_7 + sp_6
    | sp_8 <->{displace}{displace} sp_5
    | sp_5 ->{displace} sp_4 + sp_3
    | sp_3 <->{displace}{displace} sp_2

    Args:
      fh (filehandle): The function prints to this filehandle.
      ...
    
    directive simulator deterministic
    directive simulator cme
    directive stochastic { scale = 1 }

    directive simulation {
        initial=0.0; 
        final=6000.0; 
        points=1000; 
        plots=[Signal]
    }

    directive inference { 
        burnin = 1000; 
        samples = 5000; 
        thin = 100 
    }

    directive parameters [
      k=0.0003, {distribution=Uniform(1E-05, 0.002); interval=Log; variation=Random};
      bad=0.2, {distribution=Uniform(0.0, 0.3); interval=Real; variation=Random};
      T1=600.0, {distribution=Uniform(0.0, 1800.0); interval=Real;  variation=Random};
      N = 0.6;
    ]
    """

    if detailed == condensed:
        raise Exception('Choose either detailed or condensed for vdsd output.')

    if detailed:
        complexes = natural_sort(enumerator.resting_complexes) + \
                    natural_sort(enumerator.transient_complexes)
    else:
        complexes = natural_sort(enumerator.resting_macrostates)

    def logicDSD(cplx):
        """Translates a complex sequence / structure into LogicDSD notation."""
        seq = cplx.sequence
        dpp = cplx.structure
        lst = [str(d) for d in seq]
        stack, c = [], 1
        for i, (dom, pair) in enumerate(zip(seq, dpp)):
            if pair == '.':
                assert lst[i] == str(dom)
                if dom.dtype == 'short':
                    lst[i] = dom.name[:-1]+'^*' if dom.name[-1] == '*' else str(dom)+'^'
            elif pair == '+':
                assert lst[i] == '+'
                lst[i] = '> | <'
            elif pair == '(':
                assert lst[i] == str(dom)
                if dom.dtype == 'short':
                    lst[i] = dom.name[:-1]+'^*' if dom.name[-1] == '*' else str(dom)+'^'
                stack.append(i)
            elif pair == ')':
                assert lst[i] == str(dom)
                if dom.dtype == 'short':
                    lst[i] = dom.name[:-1]+'^*' if dom.name[-1] == '*' else str(dom)+'^'
                try:
                    j = stack.pop()
                except IndexError as e:
                    raise NuskellObjectError(
                        "Too many closing brackets in secondary structure")
                lst[i] += '!' + str(c)
                lst[j] += '!' + str(c)
                c += 1
            else:
                raise Exception('weird character:', pair)
        return "< {:s} >".format(' '.join(lst))

    out = []
    def output_string(string):
        if fh is None:
            out.append(string)
        else :
            fh.write(string)

    output_string("(* File autogenerated by peppercorn-{} *)\n\n".format(__version__))

    #t0, t8, tnum = 0, 10, 100
    output_string("directive simulation {\n" + 
            #"   initial={:.2f};\n".format(t0) + 
            #"   final={:d};\n".format(t8) +
            #"   points={:d};\n".format(tnum) +
            "   plots=[{:s}];\n}}\n\n".format(
                "; ".join(map(str, complexes))))
 
    #output_string("directive stochastic {scale = 1}\n\n")
    #output_string("directive simulator cme\n\n")
    
    output_string("(* LogicDSD species:\n")
    for cplx in complexes:
        if condensed: cplx = cplx.canonical
        output_string("{:s} = {:s}\n".format(cplx.name, logicDSD(cplx)))
    output_string("*)\n\n")

    if condensed :
        # Print resting macrostates
        output_string("\n(* Resting macrostate concentrations ({}) *)\n".format(len(complexes)))

        for resting in complexes:
            clist = [c.concentrationformat('nM').value \
                    for c in resting.complexes if c.concentration is not None]
            mconc = sum(clist)
            if mconc:
                output_string("| {:d} {:s}\n".format(int(mconc), resting.canonical_name))
            elif len(clist) < len(resting):
                output_string("| {:d} {:s}\n".format(1, resting.canonical_name))

        # Print reactions
        output_string("\n(* Condensed reactions ({}) *)\n".format(
            len(enumerator.condensed_reactions)))
        for rxn in natural_sort(enumerator.condensed_reactions):
            output_string("| {:s} -> {{{:g}}} {:s}\n".format(
                ' + '.join(map(str, rxn.reactants)), rxn.const, 
                ' + '.join(map(str, rxn.products))))

    else:
        # Print resting and transient complexes
        output_string("(* Initial complex concentrations ({}) *)\n".format(len(complexes)))
        for cplx in complexes:
            if cplx.concentration:
                if cplx.concentration.value:
                    output_string("| {:d} {:s}\n".format(
                        int(cplx.concentrationformat('nM').value), cplx.name))
            else:
                output_string("| {:d} {:s}\n".format(1, cplx.name))

        # Print reactions
        output_string("\n(* Detailed reactions ({}) *)\n".format(len(enumerator.reactions)))
        for rxn in natural_sort(enumerator.reactions):
            units = ['nM' if u == 'M' else u for u in rxn.rate.units]
            output_string("| {:s} -> {{{:g}}} {:s}\n".format(
                ' + '.join(map(str, rxn.reactants)), rxn.rateformat(units).constant, 
                ' + '.join(map(str, rxn.products))))

    return ''.join(out)

def write_sbml(enumerator, fh = None, condensed = False, compartment = 'TestTube'):
    """ Export reaction system to SBML Version 3 Lvl 2:

    Potentially useful links ...
        http://sbml.org/Documents/Specifications
        https://github.com/matthiaskoenig/sbmlutils
        https://sys-bio.github.io/roadrunner/python_docs/introduction.html
    """
    molarity = 'M' # as opposed to nM, uM, mM, etc; [mole / liter]
    time = 's'
    volume = 1 # 1.66 * 1e-15 # liter 

    if molarity != 'M':
        log.error('M is currently the only supported concentration format for SBML.')
        molarity = 'M'

    if time != 's':
        log.error('Seconds (s) is currently the only supported time unit for SBML.')
        time = 's'

    max_rar = 2 # Determine maximum number of reactants to define global units later ...
    reactions = enumerator.condensed_reactions if condensed else enumerator.reactions
    for rxn in reactions:
        rar = rxn.arity[0] - 1
        if rar > max_rar:
            max_rar = rar

    # -----------------
    # Helper functions:
    # -----------------
    def xml_list_of_species():
        # Returs species in counts [mole], not concentration [mole/liter]
        def xml_species(name, init, unit):
            assert unit == "mole"
            return """<species compartment="{:s}" id="{:s}" initialAmount="{:g}" 
                            boundaryCondition="false" hasOnlySubstanceUnits="false" 
                            substanceUnits="mole" constant="false"/>""".format(
                                    compartment, name, init, unit)
        xml = ''
        if condensed:
            for rm in enumerator.resting_macrostates:
                mole_list = [c.concentrationformat(molarity).value * volume \
                         for c in rm.complexes if c.concentration is not None]
                moles = sum(mole_list) # if sum(clist) else 0.0 # if else needed?
                xml += xml_species(rm.canonical_complex.name, moles, 'mole')
        else:
            for cplx in enumerator.resting_complexes + enumerator.transient_complexes:
                moles = cplx.concentrationformat(molarity).value * volume if \
                        cplx.concentration else 0.0
                xml += xml_species(cplx.name, moles, 'mole')
        return xml

    def xml_list_of_reactions():
        # Returns rate constants in /s, /M/s, /M/M/s, ... to yield rates in M/s (= mole/L/s)
        def xml_species(name, count):
            return '<speciesReference species="{:s}" stoichiometry="{:d}" constant="true"/>'.format(
                    name, count)
        def xml_reaction(rxn):
            reactants = Counter([str(s) for s in rxn.reactants])
            products = Counter([str(s) for s in rxn.products])
            reac = '\n'.join([xml_species(k,v) for k,v in reactants.items()])
            prod = '\n'.join([xml_species(k,v) for k,v in products.items()])

            rxnID = '{}__{}'.format('_'.join(reactants.elements()),'_'.join(products.elements()))
            
            law = '<apply> <times/> <ci>k</ci> {:s} </apply>'.format(
                        ' '.join(['<ci>{:s}</ci>'.format(e) for e in reactants.elements()]))

            rar = rxn.arity[0] - 1
            units = [molarity] * rar + [time]
            ratec = rxn.rateformat(units).constant # /M ... /sec
            txtunits = 'per_molar_' * rar + 'per_second'
            par = '<localParameter id="k" value="{:g}" units="{:s}"/>'.format(ratec, txtunits)

            return """
                <reaction id="{:s}" reversible="false">
                    <listOfReactants>
                        {:s}
                    </listOfReactants>
                    <listOfProducts>
                        {:s}
                    </listOfProducts>
                    <kineticLaw>
                        <math xmlns="http://www.w3.org/1998/Math/MathML">
                        <apply> <times/> <ci>{:s}</ci> 
                            {:s}
                        </apply>
                        </math>
                        <listOfLocalParameters>
                            {:s}
                        </listOfLocalParameters>
                    </kineticLaw>
                </reaction>
                """.format(rxnID, reac, prod, compartment, law, par)
        xml = ''
        reactions = enumerator.condensed_reactions if condensed else enumerator.reactions
        for rxn in reactions:
            xml += xml_reaction(rxn)
        return xml

    def xml_list_of_units(max_reactants):
        def unit_definition(rar):
            txtunits = 'per_molar_' * rar + 'per_second'
            return """
                    <unitDefinition id="{:s}">
                        <listOfUnits>
                        <unit kind="mole" exponent="-{:d}" scale="0" multiplier="1"/>
                        <unit kind="litre" exponent="{:d}" scale="0" multiplier="1"/>
                        <unit kind="second" exponent="-1" scale="0" multiplier="1"/>
                        </listOfUnits>
                    </unitDefinition>
                    """.format(txtunits, rar, rar)
        xml = ''
        for e in range(1, max_reactants + 1):
            xml += unit_definition(e)
        return xml

    # Ok, so let's set this up ...
    xmlspex = """
    <sbml xmlns="http://www.sbml.org/sbml/level3/version2/core" level="3" version="2">
    <model extentUnits="mole" timeUnits="second">
        <listOfUnitDefinitions>
            <unitDefinition id="per_second">
                <listOfUnits>
                    <unit kind="second" exponent="-1" scale="0" multiplier="1"/>
                </listOfUnits>
            </unitDefinition>
            {:s}
        </listOfUnitDefinitions>
        <listOfCompartments>
            <compartment id="{:s}" 
                size="{:g}" spatialDimensions="3" units="litre" constant="true"/>
        </listOfCompartments>
        <listOfSpecies>
            {:s}
        </listOfSpecies>
        <listOfReactions>
            {:s}
        </listOfReactions>
    </model>
    </sbml>
    """.format(xml_list_of_units(max_rar), compartment, volume, xml_list_of_species(), xml_list_of_reactions())

    doc = xml.dom.minidom.parseString(xmlspex)
    doc = doc.toprettyxml(indent = ' ', encoding="UTF-8")
    doc = '\n'.join([s.decode() for s in doc.splitlines() if s.strip()]) + '\n'

    if fh:
        fh.write(doc)
    return '' if fh else doc

